import numpy as np
from ase.units import Hartree
from Scientific.IO.NetCDF import NetCDFFile


class ETSFWriter:
    def __init__(self, filename='gpaw', title='gpaw'):
        if not filename.endswith('-etsf.nc'):
            if filename.endswith('.nc'):
                filename = filename[:-3] + '-etsf.nc'
            else:
                filename = filename + '-etsf.nc'
            
        self.nc = NetCDFFile(filename, 'w')

        self.nc.file_format = 'ETSF Nanoquanta'
        self.nc.file_format_version = np.array([3.3], dtype=np.float32)
        self.nc.Conventions = 'http://www.etsf.eu/fileformats/'
        self.nc.history = 'File generated by GPAW'
        self.nc.title = title

    def write(self, calc, spacegroup=1):

        #sg = Spacegroup(spacegroup)
        #print sg
        
        wfs = calc.wfs
        setups = wfs.setups
        bd = wfs.bd
        kd = wfs.kd
        
        atoms = calc.atoms
        natoms = len(atoms)
        
        if wfs.kd.symmetry is None:
            op_scc = np.eye(3, dtype=int).reshape((1, 3, 3))
        else:
            op_scc = wfs.kd.symmetry.op_scc

        specie_a = np.empty(natoms, np.int32)
        nspecies = 0
        species = {}
        names = []
        symbols = []
        numbers = []
        charges = []
        for a, id in enumerate(setups.id_a):
            if id not in species:
                species[id] = nspecies
                nspecies += 1
                names.append(setups[a].symbol)
                symbols.append(setups[a].symbol)
                numbers.append(setups[a].Z)
                charges.append(setups[a].Nv)
            specie_a[a] = species[id]
            
        dimensions = [
            ('character_string_length', 80),
            ('max_number_of_states', bd.nbands),
            ('number_of_atoms', len(atoms)),
            ('number_of_atom_species', nspecies),
            ('number_of_cartesian_directions', 3),
            ('number_of_components', 1),
            ('number_of_kpoints', kd.nibzkpts),
            ('number_of_reduced_dimensions', 3),
            ('number_of_spinor_components', 1),
            ('number_of_spins', wfs.nspins),
            ('number_of_symmetry_operations', len(op_scc)),
            ('number_of_vectors', 3),
            ('real_or_complex_coefficients', 2),
            ('symbol_length', 2)]

        for name, size in dimensions:
            print(('%-34s %d' % (name, size)))
            self.nc.createDimension(name, size)

        var = self.add_variable
        
        var('space_group', (), np.array(spacegroup, dtype=int))
        var('primitive_vectors',
            ('number_of_vectors', 'number_of_cartesian_directions'),
            wfs.gd.cell_cv, units='atomic units')
        var('reduced_symmetry_matrices',
            ('number_of_symmetry_operations',
             'number_of_reduced_dimensions', 'number_of_reduced_dimensions'),
            op_scc.astype(np.int32), symmorphic='yes')
        var('reduced_symmetry_translations',
            ('number_of_symmetry_operations', 'number_of_reduced_dimensions'),
            np.zeros((len(op_scc), 3), dtype=np.int32))
        var('atom_species', ('number_of_atoms',), specie_a + 1)
        var('reduced_atom_positions',
            ('number_of_atoms', 'number_of_reduced_dimensions'),
            calc.spos_ac)
        var('atomic_numbers', ('number_of_atom_species',),
            np.array(numbers, dtype=float))
        var('valence_charges', ('number_of_atom_species',),
            np.array(charges, dtype=float))
        var('atom_species_names',
            ('number_of_atom_species', 'character_string_length'), names)
        var('chemical_symbols', ('number_of_atom_species', 'symbol_length'),
            symbols)
        var('pseudopotential_types',
            ('number_of_atom_species', 'character_string_length'),
            ['HGH'] * nspecies)
        var('fermi_energy', (), calc.occupations.fermilevel,
            units='atomic units')
        var('smearing_scheme', ('character_string_length',), 'fermi-dirac')
        var('smearing_width', (), calc.occupations.width, units='atomic units')
        var('number_of_states', ('number_of_spins', 'number_of_kpoints'),
            np.zeros((wfs.nspins, kd.nibzkpts), np.int32) + bd.nbands,
            k_dependent='no')
        var('eigenvalues',
            ('number_of_spins', 'number_of_kpoints', 'max_number_of_states'),
            np.array([[calc.get_eigenvalues(k, s) / Hartree
                       for k in range(kd.nibzkpts)]
                      for s in range(wfs.nspins)]), units='atomic units')
        var('occupations',
            ('number_of_spins', 'number_of_kpoints', 'max_number_of_states'),
            np.array([[calc.get_occupation_numbers(k, s) / kd.weight_k[k]
                       for k in range(kd.nibzkpts)]
                      for s in range(wfs.nspins)]))
        var('reduced_coordinates_of_kpoints',
            ('number_of_kpoints', 'number_of_reduced_dimensions'), kd.ibzk_kc)
        var('kpoint_weights', ('number_of_kpoints',), kd.weight_k)
        var('basis_set', ('character_string_length',), 'plane_waves')
        var('number_of_electrons', (), np.array(wfs.nvalence, dtype=np.int32))
        self.nc.close()

    def add_variable(self, name, dims, data=None, **kwargs):
        if data is None:
            char = 'd'
        else:
            if isinstance(data, np.ndarray):
                char = data.dtype.char
            elif isinstance(data, float):
                char = 'd'
            elif isinstance(data, int):
                char = 'i'
            else:
                char = 'c'
        print(('%-34s %s%s' % (
            name, char,
            tuple([self.nc.dimensions[dim] for dim in dims]))))
        var = self.nc.createVariable(name, char, dims)
        for attr, value in kwargs.items():
            setattr(var, attr, value)
        if data is not None:
            if len(dims) == 0:
                var.assignValue(data)
            else:
                if char == 'c':
                    if len(dims) == 1:
                        var[:len(data)] = data
                    else:
                        for i, x in enumerate(data):
                            var[i, :len(x)] = x
                else:
                    var[:] = data
        return var



from __future__ import print_function
from ase import Atoms
from gpaw import GPAW
from gpaw.test import equal

s = Atoms('Cl')
s.center(vacuum=3)
c = GPAW(xc={'name': 'PBE', 'stencil': 1}, nbands=-4, charge=-1, h=0.3)
s.set_calculator(c)

e = s.get_potential_energy()
niter = c.get_number_of_iterations()

print(e, niter)
energy_tolerance = 0.004
equal(e, -2.8967, energy_tolerance)



import numpy as np
from ase import Atoms
from gpaw import GPAW
from gpaw.wavefunctions.pw import PW
from gpaw.test import equal

bulk = Atoms('Li', pbc=True)
k = 4
calc = GPAW(mode=PW(200),
            kpts=(k, k, k),
            eigensolver='rmm-diis')

bulk.set_calculator(calc)
e = []
niter = []
A = [2.6, 2.65, 2.7, 2.75, 2.8]
for a in A:
    bulk.set_cell((a, a, a))
    e.append(bulk.get_potential_energy())

a = np.roots(np.polyder(np.polyfit(A, e, 2), 1))[0]
print('a =', a)
equal(a, 2.65247379609, 0.001)



from __future__ import print_function

from ase import Atom, Atoms

from gpaw import GPAW
from gpaw.test import equal

try:
    calc = GPAW('NaCl.gpw')
    NaCl = calc.get_atoms()
    e = NaCl.get_potential_energy()
    niter = None
except IOError:
    h = 0.21  # gridspacing
    a = [6.5, 6.5, 7.7]  # unit cell
    d = 2.3608  # experimental bond length

    NaCl = Atoms([Atom('Na', [0, 0, 0]),
                  Atom('Cl', [0, 0, d])],
                 pbc=False, cell=a)
    NaCl.center()
    calc = GPAW(h=h, xc='LDA', nbands=5,
                setups={'Na': '1'},
                convergence={'eigenstates': 1e-6}, spinpol=1)

    NaCl.set_calculator(calc)
    e = NaCl.get_potential_energy()
    niter = calc.get_number_of_iterations()
    calc.write('NaCl.gpw')

equal(e, -4.907, 0.002)

dv = NaCl.get_volume() / calc.get_number_of_grid_points().prod()
nt1 = calc.get_pseudo_density(gridrefinement=1)
Zt1 = nt1.sum() * dv
nt2 = calc.get_pseudo_density(gridrefinement=2)
Zt2 = nt2.sum() * dv / 8
print('Integral of pseudo density:', Zt1, Zt2)
equal(Zt1, Zt2, 1e-12)

for gridrefinement in [1, 2, 4]:
    n = calc.get_all_electron_density(gridrefinement=gridrefinement)
    Z = n.sum() * dv / gridrefinement**3
    print('Integral of all-electron density:', Z)
    equal(Z, 28, 1e-5)



import numpy as np

from ase import Atoms
from ase.parallel import paropen

from gpaw import GPAW, FermiDirac, Davidson
from gpaw.wavefunctions.pw import PW
from gpaw.response.chi0 import Chi0
from gpaw.test import equal

# This script asserts that the chi's obtained
# from GS calculations using symmetries
# and GS calculations not using symmetries return
# the same results. Tests that the chi's are element-wise
# equal to a tolerance of 1e-10.

resultfile = paropen('results.txt', 'a')
pwcutoff = 400.0
k = 4
a = 4.59
c = 2.96
u = 0.305

rutile_cell = [[a, 0, 0],
               [0, a, 0],
               [0, 0, c]]

TiO2_basis = np.array([[0.0, 0.0, 0.0],
                       [0.5, 0.5, 0.5],
                       [u, u, 0.0],
                       [-u, -u, 0.0],
                       [0.5 + u, 0.5 - u, 0.5],
                       [0.5 - u, 0.5 + u, 0.5]])

bulk_crystal = Atoms(symbols=['Ti', 'Ti', 'O', 'O', 'O', 'O'],
                     scaled_positions=TiO2_basis,
                     cell=rutile_cell,
                     pbc=(1, 1, 1))
data_s = []
for symmetry in ['off', {}]:
    bulk_calc = GPAW(mode=PW(pwcutoff),
                     nbands=42,
                     eigensolver=Davidson(1),
                     kpts={'size': (k, k, k), 'gamma': True},
                     xc='PBE',
                     occupations=FermiDirac(0.00001),
                     parallel={'band': 1},
                     symmetry=symmetry)

    bulk_crystal.set_calculator(bulk_calc)
    e0_bulk_pbe = bulk_crystal.get_potential_energy()
    bulk_calc.write('bulk.gpw', mode='all')
    X = Chi0('bulk.gpw')
    chi_t = X.calculate([1. / 4, 0, 0])[1:]
    data_s.append(list(chi_t))

msg = 'Difference in Chi when turning off symmetries!'

while len(data_s):
    data1 = data_s.pop()
    for data2 in data_s:
        for dat1, dat2 in zip(data1, data2):
            if dat1 is not None:
                equal(np.abs(dat1 - dat2).max(),
                      0, 0.001, msg=msg)



from __future__ import print_function
from ase.build import molecule
from gpaw import GPAW, Mixer, Davidson, PW
from gpaw.xc.libvdwxc import vdw_df
from gpaw.mpi import world

system = molecule('H2')
system.center(vacuum=1.0)
system.pbc = 1

def calculate(**kwargs1):
    kwargs = dict(mode=mode,
                  basis='sz(dzp)',
                  eigensolver=Davidson(4) if mode != 'lcao' else None,
                  xc=vdw_df(),
                  h=0.25,
                  convergence=dict(energy=1e-6),
                  mixer=Mixer(0.5, 5, 10.))
    kwargs.update(kwargs1)
    calc = GPAW(**kwargs)
    system.set_calculator(calc)
    system.get_potential_energy()
    return calc

fddomainpar = min(2, world.size)

for mode in ['fd', 'pw', 'lcao']:
    kwargs = {}
    if mode == 'pw':
        kwargs['mode'] = PW(250)
    else:
        kwargs['parallel'] = {'domain': fddomainpar}
    calc = calculate(**kwargs)

E1 = calc.get_potential_energy()
calc.write('dump.libvdwxc.gpw')
calc2 = GPAW('dump.libvdwxc.gpw', txt='restart.txt',
             parallel={'domain': fddomainpar})
system2 = calc.get_atoms()

# Verify same energy after restart
E2 = system2.get_potential_energy()
assert abs(E2 - E1) < 1e-14  # Should be exact

# Trigger recaclulation of essentially same system
system2.positions[0, 0] += 1e-13
print('reconverge')
E3 = system2.get_potential_energy()
err2 = abs(E3 - E2)
print('error', err2)
assert err2 < 1e-6  # Around SCF precision



class NonLocalFunctionalFactory:
    """Factory class.

    NonLocalFunctionalFactory is used by xc_functional.py, when the
    functional starts with words GLLB.

    It contains a method called get_functional_by_name, which takes
    the xc-name for non-local functional and returns the corresponding
    NonLocalFunctional object.

    * GLLB
    * GLLBC (GLLB with screening part from PBE + PBE Correlation)
    * GLLBSC (GLLB with screening part from PBE_SOL + PBE Correlation)


    * GLLBNORESP (Just GLLB Screening)
    * GLLBLDA (A test functional, which is just LDA but via
               NonLocalFunctional framework)
    * GLLBGGA (A test functional, which is just GGA but via
               NonLocalFunctional framework)
    """

    def get_functional_by_name(self, name):
        from gpaw.xc.gllb.nonlocalfunctional import NonLocalFunctional
        functional = NonLocalFunctional(name)

        if name == 'GLLB':
            # Functional GLLB
            # Contains screening part from GGA functional
            # And response part based on simple square root expection
            # of orbital energy differences.
            from gpaw.xc.gllb.c_gllbscr import C_GLLBScr
            from gpaw.xc.gllb.c_response import C_Response
            C_Response(functional, 1.0,
                       C_GLLBScr(functional, 1.0).get_coefficient_calculator())
            return functional
        elif name == 'GLLBM':
            from gpaw.xc.gllb.c_gllbscr import C_GLLBScr
            from gpaw.xc.gllb.c_response import C_Response
            from gpaw.xc.gllb.c_xc import C_XC
            C_Response(functional, 1.0, C_GLLBScr(
                functional, 1.0, metallic=True).get_coefficient_calculator())
            return functional
        elif name == 'GLLBSC':
            from gpaw.xc.gllb.c_gllbscr import C_GLLBScr
            from gpaw.xc.gllb.c_response import C_Response
            from gpaw.xc.gllb.c_xc import C_XC
            C_Response(functional, 1.0, C_GLLBScr(
                functional,
                1.0,
                'GGA_X_PBE_SOL').get_coefficient_calculator())
            C_XC(functional, 1.0, 'GGA_C_PBE_SOL')
            return functional
        elif name == 'GLLBSCM':
            from gpaw.xc.gllb.c_gllbscr import C_GLLBScr
            from gpaw.xc.gllb.c_response import C_Response
            from gpaw.xc.gllb.c_xc import C_XC
            C_Response(functional, 1.0, C_GLLBScr(
                functional,
                1.0,
                'GGA_X_PBE_SOL', metallic=True).get_coefficient_calculator())
            C_XC(functional, 1.0, 'GGA_C_PBE_SOL')
            return functional

        elif name == 'GLLBC':
            from gpaw.xc.gllb.c_gllbscr import C_GLLBScr
            from gpaw.xc.gllb.c_response import C_Response
            from gpaw.xc.gllb.c_xc import C_XC
            C_Response(functional, 1.0,
                       C_GLLBScr(functional, 1.0,'X_PBE-None').get_coefficient_calculator())
            C_XC(functional, 1.0, 'None-C_PBE')
            return functional
        elif name == 'GLLBCP86':
            from gpaw.xc.gllb.c_gllbscr import C_GLLBScr
            from gpaw.xc.gllb.c_response import C_Response
            from gpaw.xc.gllb.c_xc import C_XC
            C_Response(functional, 1.0,
                       C_GLLBScr(functional, 1.0).get_coefficient_calculator())
            C_XC(functional, 1.0, 'None-C_P86')
            return functional
        elif name == 'GLLBLDA':
            from gpaw.xc.gllb.c_xc import C_XC
            C_XC(functional, 1.0,'LDA')
            return functional
        elif name == 'GLLBGGA':
            from gpaw.xc.gllb.c_xc import C_XC
            C_XC(functional, 1.0,'PBE')
            return functional
        elif name == 'GLLBSLATER':
            from gpaw.xc.gllb.c_slater import C_Slater
            C_Slater(functional, 1.0)
            return functional
        elif name == 'GLLBNORESP':
            from gpaw.xc.gllb.c_gllbscr import C_GLLBScr
            C_GLLBScr(functional, 1.0)
            return functional
        elif name == 'KLI':
            raise RuntimeError('KLI functional not implemented')
            from gpaw.xc.gllb.c_slater import C_Slater
            from gpaw.xc.gllb.c_response import C_Response
            C_Response(functional, 1.0,
                       C_Slater(functional, 1.0).get_coefficient_calculator())
            return functional
        else:
            raise RuntimeError('Unkown NonLocal density functional: ' + name)



import sys
import time

from ase import Atom, Atoms
from ase.parallel import parprint

from gpaw import GPAW
from gpaw.test import equal
from gpaw.lrtddft import LrTDDFT
from gpaw.lrtddft.excited_state import ExcitedState

txt='-'
txt='/dev/null'

R=0.7 # approx. experimental bond length
a = 3.0
c = 4.0
H2 = Atoms([Atom('H', (a / 2, a / 2, (c - R) / 2)),
            Atom('H', (a / 2, a / 2, (c + R) / 2))],
           cell=(a, a, c))
calc = GPAW(xc='PBE', h=0.25, nbands=3, spinpol=False, txt=txt)
H2.set_calculator(calc)

xc='LDA'
lr = LrTDDFT(calc, xc=xc)

# excited state with forces
accuracy = 0.015
exst = ExcitedState(lr, 0, d=0.01,
                    parallel=2,
                    txt=sys.stdout)

t0 = time.time()
parprint("########### first call to forces --> calculate")
forces = exst.get_forces(H2)
parprint("time used:", time.time() - t0)
for c in range(2):
    equal(forces[0,c], 0.0, accuracy)
    equal(forces[1,c], 0.0, accuracy)
equal(forces[0, 2] + forces[1, 2], 0.0, accuracy)


parprint("########### second call to potential energy --> just return")
t0 = time.time()
E = exst.get_potential_energy()
parprint("E=", E)
parprint("time used:", time.time() - t0)
t0 = time.time()
E = exst.get_potential_energy(H2)
parprint("E=", E)
parprint("time used:", time.time() - t0)

parprint("########### second call to forces --> just return")
t0 = time.time()
exst.get_forces()
parprint("time used:", time.time() - t0)
t0 = time.time()
exst.get_forces(H2)
parprint("time used:", time.time() - t0)

parprint("###########  moved atoms, call to forces --> calculate")
p = H2.get_positions()
p[1, 1] += 0.1
H2.set_positions(p)

t0 = time.time()
exst.get_forces(H2)
parprint("time used:", time.time() - t0)



from __future__ import print_function

import numpy as np
from ase import Atoms

from gpaw import GPAW
from gpaw.test import equal
from gpaw.wannier import Wannier

# GPAW wannier example for ethylene corresponding to the ASE Wannier
# tutorial.

a = 6.0  # Size of unit cell (Angstrom)

ethylene = Atoms('H2C2H2',
                 [(-1.235, -0.936, 0),
                  (-1.235, 0.936, 0),
                  (-0.660, 0.000, 0),
                  (0.660, 0.000, 0),
                  (1.235, -0.936, 0),
                  (1.235, 0.936, 0)],
                 cell=(a, a, a),
                 pbc=True)
ethylene.center()

calc = GPAW(nbands=8, gpts=(32, 32, 32), convergence={'eigenstates': 3.3e-5})
ethylene.set_calculator(calc)
e = ethylene.get_potential_energy()
niter = calc.get_number_of_iterations()

energy_tolerance = 0.002
niter_tolerance = 0
equal(e, -33.328, energy_tolerance)


def check(calc):
    wannier = Wannier(calc, nbands=6)
    wannier.localize()

    centers = wannier.get_centers()
    print(centers)
    expected = [[1.950, 2.376, 3.000],
                [1.950, 3.624, 3.000],
                [3.000, 3.000, 2.671],
                [3.000, 3.000, 3.329],
                [4.050, 2.376, 3.000],
                [4.050, 3.624, 3.000]]
    equal(13.7995, wannier.value, 0.016)
    for center in centers:
        i = 0
        while np.sum((expected[i] - center)**2) > 0.01:
            i += 1
            if i == len(expected):
                raise RuntimeError('Correct center not found')
    expected.pop(i)


check(calc)
calc.write('ethylene.gpw', 'all')
check(GPAW('ethylene.gpw', txt=None))

try:
    from gpaw.io.etsf import ETSFWriter
except ImportError:
    pass  # Scientific.IO.NetCDF was not installed
else:
    if calc.wfs.world.size == 1:
        ETSFWriter().write(calc)

# for i in range(6):
#     wannier.write_cube(i, 'ethylene%s.cube' % i, real=True)

# from ASE.Visualization.PrimiPlotter import PrimiPlotter, X11Window
# ethylene.extend(wannier.get_centers_as_atoms())
# plot = PrimiPlotter(ethylene)
# plot.set_output(X11Window())
# plot.set_radii(.2)
# plot.set_rotation([15, 0, 0])
# plot.plot()



from __future__ import division
from ase import Atoms
from ase.lattice.hexagonal import Hexagonal
from gpaw import GPAW, FermiDirac
from gpaw.test import equal
from gpaw.response.g0w0 import G0W0

if 1:
    calc = GPAW(mode='pw',
                xc='PBE',
                experimental={'niter_fixdensity': 2},
                nbands=16,
                setups={'Mo': '6'},
                eigensolver='rmm-diis',
                occupations=FermiDirac(0.001),
                kpts={'size': (6, 6, 1), 'gamma': True})
    
    a = 3.1604
    c = 10.0
    
    cell = Hexagonal(symbol='Mo', latticeconstant={'a': a, 'c': c}).get_cell()
    layer = Atoms(symbols='MoS2', cell=cell, pbc=True,
                  scaled_positions=[(0, 0, 0),
                                    (2 / 3, 1 / 3, 0.3),
                                    (2 / 3, 1 / 3, -0.3)])
    
    pos = layer.get_positions()
    pos[1][2] = pos[0][2] + 3.172 / 2
    pos[2][2] = pos[0][2] - 3.172 / 2
    layer.set_positions(pos)
    layer.set_calculator(calc)
    layer.get_potential_energy()
    calc.write('MoS2.gpw', mode='all')

gw = G0W0('MoS2.gpw',
          'gw-test',
          nbands=15,
          ecut=10,
          domega0=0.1,
          eta=0.2,
          truncation='2D',
          kpts=[((1 / 3, 1 / 3, 0))],
          bands=(8, 10),
          savepckl=True)

e_qp = gw.calculate()['qp'][0, 0]

ev = 2.669
ec = 6.831
equal(e_qp[0], ev, 0.01)
equal(e_qp[1], ec, 0.01)



class CLICommand:
    """Run GPAW's Python interpreter."""

    @staticmethod
    def add_arguments(parser):
        parser.add_argument('arguments', nargs='*')

    @staticmethod
    def run(args):
        print('Use: gpaw -P <nproc> python -- <arguments>')



import numpy as np

from ase.lattice.hexagonal import Graphene

from gpaw import GPAW, FermiDirac
from gpaw.test import equal
from gpaw.response.g0w0 import G0W0

""" Tests extrapolation to infinite energy cutoff + block paralellization.
It takes ~109 s on one core"""

if 1:
    calc = GPAW(mode='pw',
                xc='PBE',
                nbands=16,
                setups={'Mo': '6'},
                eigensolver='rmm-diis',
                occupations=FermiDirac(0.001),
                parallel={'domain': 1},
                kpts={'size': (6, 6, 1), 'gamma': True})

    layer = Graphene(symbol='B',
                     latticeconstant={'a': 2.5, 'c': 1.0},
                     size=(1, 1, 1))
    layer[0].symbol = 'N'
    layer.pbc = (1, 1, 0)
    layer.center(axis=2, vacuum=4.0)

    layer.set_calculator(calc)
    layer.get_potential_energy()

    nbecut = 50
    from ase.units import Bohr, Hartree
    vol = layer.get_volume() / Bohr**3
    nbands = int(vol * (nbecut / Hartree)**1.5 * 2**0.5 / 3 / np.pi**2)
    calc.diagonalize_full_hamiltonian(nbands)
    calc.write('hBN.gpw', mode='all')

gw = G0W0('hBN.gpw',
          'gw-hBN',
          ecut=50,
          domega0=0.1,
          eta=0.2,
          truncation='2D',
          kpts=[0],
          bands=(3, 5),
          ecut_extrapolation=[30, 40, 50],
          nblocksmax=True)

e_qp = gw.calculate()['qp'][0, 0]

ev = -4.381
ec = 3.71013806
equal(e_qp[0], ev, 0.01)
equal(e_qp[1], ec, 0.01)



from __future__ import print_function
import os
import sys
from time import time

import numpy as np
from scipy.special import sici
from scipy.special.orthogonal import p_roots
import ase.io.ulm as ulm
from ase.units import Ha, Bohr
from ase.utils.timing import timer

import gpaw.mpi as mpi
from gpaw.blacs import BlacsGrid, Redistributor
from gpaw.fd_operators import Gradient
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.utilities.blas import gemmdot, axpy
from gpaw.wavefunctions.pw import PWDescriptor
from gpaw.xc.rpa import RPACorrelation

class FXCCorrelation(RPACorrelation):
    def __init__(self, calc, xc='RPA', filename=None,
                 skip_gamma=False, qsym=True, nlambda=8,
                 nfrequencies=16, frequency_max=800.0, frequency_scale=2.0,
                 frequencies=None, weights=None, density_cut=1.e-6,
                 world=mpi.world, nblocks=1,
                 unit_cells=None, tag=None,
                 txt=sys.stdout, range_rc=1.0, av_scheme=None, Eg=None):

        RPACorrelation.__init__(self, calc, xc=xc, filename=filename,
                                skip_gamma=skip_gamma, qsym=qsym,
                                nfrequencies=nfrequencies, nlambda=nlambda,
                                frequency_max=frequency_max,
                                frequency_scale=frequency_scale,
                                frequencies=frequencies, weights=weights,
                                world=world, nblocks=nblocks,
                                txt=txt)

        self.l_l, self.weight_l = p_roots(nlambda)
        self.l_l = (self.l_l + 1.0) * 0.5
        self.weight_l *= 0.5
        self.xc = xc
        self.density_cut = density_cut
        if unit_cells is None:
            unit_cells = self.calc.wfs.kd.N_c
        self.unit_cells = unit_cells
        self.range_rc = range_rc    # Range separation parameter in Bohr
        self.av_scheme = av_scheme  # Either 'density' or 'wavevector'
        self.Eg = Eg                # Band gap in eV

        set_flags(self)

        if tag is None:

            tag = self.calc.atoms.get_chemical_formula(mode='hill')

            if self.av_scheme is not None:

                tag += '_' + self.av_scheme

        self.tag = tag

    @timer('FXC')
    def calculate(self, ecut, nbands=None):

        if self.xc not in ('RPA', 'range_RPA'):
            # kernel not required for RPA/range_sep RPA

            if isinstance(ecut, (float, int)):
                self.ecut_max = ecut
            else:
                self.ecut_max = max(ecut)

            # Find the first q vector to calculate kernel for
            # (density averaging scheme always calculates all q points anyway)

            q_empty = None

            for iq in reversed(range(len(self.ibzq_qc))):

                if not os.path.isfile('fhxc_%s_%s_%s_%s.ulm'
                                      % (self.tag, self.xc,
                                         self.ecut_max, iq)):
                    q_empty = iq

            if q_empty is not None:

                if self.av_scheme == 'wavevector':

                    print('Calculating %s kernel starting from q point %s' %
                          (self.xc, q_empty), file=self.fd)
                    print(file=self.fd)

                    if self.linear_kernel:
                        kernel = KernelWave(self.calc,
                                            self.xc,
                                            self.ibzq_qc,
                                            self.fd,
                                            None,
                                            q_empty,
                                            None,
                                            self.Eg,
                                            self.ecut_max,
                                            self.tag,
                                            self.timer)

                    elif not self.dyn_kernel:
                        kernel = KernelWave(self.calc,
                                            self.xc,
                                            self.ibzq_qc,
                                            self.fd,
                                            self.l_l,
                                            q_empty,
                                            None,
                                            self.Eg,
                                            self.ecut_max,
                                            self.tag,
                                            self.timer)

                    else:
                        kernel = KernelWave(self.calc,
                                            self.xc,
                                            self.ibzq_qc,
                                            self.fd,
                                            self.l_l,
                                            q_empty,
                                            self.omega_w,
                                            self.Eg,
                                            self.ecut_max,
                                            self.tag,
                                            self.timer)

                else:

                    kernel = KernelDens(self.calc,
                                        self.xc,
                                        self.ibzq_qc,
                                        self.fd,
                                        self.unit_cells,
                                        self.density_cut,
                                        self.ecut_max,
                                        self.tag,
                                        self.timer)

                kernel.calculate_fhxc()
                del kernel

            else:
                print('%s kernel already calculated' % self.xc, file=self.fd)
                print(file=self.fd)

        if self.xc in ('range_RPA', 'range_rALDA'):

            shortrange = range_separated(self.calc, self.fd, self.omega_w,
                                         self.weight_w, self.l_l,
                                         self.weight_l,
                                         self.range_rc, self.xc)

            self.shortrange = shortrange.calculate()

        if self.calc.wfs.nspins == 1:
            spin = False
        else:
            spin = True

        e = RPACorrelation.calculate(self, ecut, spin=spin, nbands=nbands)

        return e

    @timer('Chi0(q)')
    def calculate_q(self, chi0, pd,
                    chi0_swGG, chi0_swxvG, chi0_swvv,
                    m1, m2, cut_G, A2_x):
        if chi0_swxvG is None:
            chi0_swxvG = range(2)  # Not used
            chi0_swvv = range(2)  # Not used
        chi0._calculate(pd, chi0_swGG[0], chi0_swxvG[0], chi0_swvv[0],
                        m1, m2, [0], extend_head=False)
        if len(chi0_swGG) == 2:
            chi0._calculate(pd, chi0_swGG[1], chi0_swxvG[1], chi0_swvv[1],
                            m1, m2, [1], extend_head=False)
        print('E_c(q) = ', end='', file=self.fd)

        if self.nblocks > 1:
            if len(chi0_swGG) == 2:
                chi0_0wGG = chi0.redistribute(chi0_swGG[0], A2_x)
                nredist = np.product(chi0_0wGG.shape)
                chi0.redistribute(chi0_swGG[1], A2_x[nredist:])
                chi0_swGG = A2_x[:2 * nredist].reshape((2,) + chi0_0wGG.shape)
                chi0_swGG = np.swapaxes(chi0_swGG,2,3)
            else:
                chi0_0wGG = chi0.redistribute(chi0_swGG[0], A2_x)
                nredist = np.product(chi0_0wGG.shape)
                chi0_swGG = A2_x[:1 * nredist].reshape((1,) + chi0_0wGG.shape)
                chi0_swGG = np.swapaxes(chi0_swGG,2,3)

        if not pd.kd.gamma:
            e = self.calculate_energy(pd, chi0_swGG, cut_G)
            print('%.3f eV' % (e * Ha), file=self.fd)
            self.fd.flush()
        else:
            nw = len(self.omega_w)
            mynw = nw // self.nblocks
            w1 = self.blockcomm.rank * mynw
            w2 = w1 + mynw
            e = 0.0
            for v in range(3):
                chi0_swGG[:, :, 0] = chi0_swxvG[:, w1:w2, 0, v]
                chi0_swGG[:, :, :, 0] = chi0_swxvG[:, w1:w2, 1, v]
                chi0_swGG[:, :, 0, 0] = chi0_swvv[:, w1:w2, v, v]
                ev = self.calculate_energy(pd, chi0_swGG, cut_G)
                e += ev
                print('%.3f' % (ev * Ha), end='', file=self.fd)
                if v < 2:
                    print('/', end='', file=self.fd)
                else:
                    print('eV', file=self.fd)
                    self.fd.flush()
            e /= 3

        return e

    @timer('Energy')
    def calculate_energy(self, pd, chi0_swGG, cut_G):
        """Evaluate correlation energy from chi0 and the kernel fhxc"""

        ibzq2_q = [np.dot(self.ibzq_qc[i] - pd.kd.bzk_kc[0],
                          self.ibzq_qc[i] - pd.kd.bzk_kc[0])
                   for i in range(len(self.ibzq_qc))]

        qi = np.argsort(ibzq2_q)[0]

        G_G = pd.G2_qG[0]**0.5  # |G+q|

        if cut_G is not None:
            G_G = G_G[cut_G]

        nG = len(G_G)
        ns = len(chi0_swGG)

        # There are three options to calculate the
        # energy depending on kernel and/or averaging scheme.

        # Option (1) - Spin-polarized form of kernel exists
        #              e.g. rALDA, rAPBE.
        #              Then, solve block diagonal form of Dyson
        #              equation (dimensions (ns*nG) * (ns*nG))
        #              (note this does not necessarily mean that
        #              the calculation is spin-polarized!)

        if self.spin_kernel:
            r = ulm.open('fhxc_%s_%s_%s_%s.ulm' %
                         (self.tag, self.xc, self.ecut_max, qi))
            fv = r.fhxc_sGsG

            if cut_G is not None:
                cut_sG = np.tile(cut_G, ns)
                cut_sG[len(cut_G):] += len(fv) // ns
                fv = fv.take(cut_sG, 0).take(cut_sG, 1)

            # the spin-polarized kernel constructed from wavevector average
            # is already multiplied by |q+G| |q+G'|/4pi, and doesn't require
            # special treatment of the head and wings.  However not true for
            # density average:

            if self.av_scheme == 'density':
                for s1 in range(ns):
                    for s2 in range(ns):
                        m1 = s1 * nG
                        n1 = (s1 + 1) * nG
                        m2 = s2 * nG
                        n2 = (s2 + 1) * nG
                        fv[m1:n1, m2:n2] *= (G_G * G_G[:, np.newaxis] /
                                             (4 * np.pi))

                        if np.prod(self.unit_cells) > 1 and pd.kd.gamma:
                            m1 = s1 * nG
                            n1 = (s1 + 1) * nG
                            m2 = s2 * nG
                            n2 = (s2 + 1) * nG
                            fv[m1, m2:n2] = 0.0
                            fv[m1:n1, m2] = 0.0
                            fv[m1, m2] = 1.0

            if pd.kd.gamma:
                G_G[0] = 1.0

            e_w = []

            # Loop over frequencies
            for chi0_sGG in np.swapaxes(chi0_swGG, 0, 1):
                if cut_G is not None:
                    chi0_sGG = chi0_sGG.take(cut_G, 1).take(cut_G, 2)
                chi0v = np.zeros((ns * nG, ns * nG), dtype=complex)
                for s in range(ns):
                    m = s * nG
                    n = (s + 1) * nG
                    chi0v[m:n, m:n] = chi0_sGG[s] / G_G / G_G[:, np.newaxis]
                chi0v *= 4 * np.pi

                del chi0_sGG

                e = 0.0

                for l, weight in zip(self.l_l, self.weight_l):
                    chiv = np.linalg.solve(np.eye(nG * ns) -
                                           l * np.dot(chi0v, fv),
                                           chi0v).real  # this is SO slow
                    for s1 in range(ns):
                        for s2 in range(ns):
                            m1 = s1 * nG
                            n1 = (s1 + 1) * nG
                            m2 = s2 * nG
                            n2 = (s2 + 1) * nG
                            chiv_s1s2 = chiv[m1:n1, m2:n2]
                            e -= np.trace(chiv_s1s2) * weight

                e += np.trace(chi0v.real)

                e_w.append(e)

        else:
            # Or, if kernel does not have a spin polarized form,
            #
            # Option (2)  kernel does not scale linearly with lambda,
            #             so we solve nG*nG Dyson equation at each value
            #             of l.  Requires kernel to be constructed
            #             at individual values of lambda
            #
            # Option (3)  Divide correlation energy into
            #             long range part which can be integrated
            #             analytically w.r.t. lambda, and a short
            #             range part which again requires
            #             solving Dyson equation (hence no speedup,
            #             but the maths looks nice and
            #             fits with range-separated RPA)
            #
            #
            # Construct/read kernels

            if self.xc == 'RPA':

                fv = np.eye(nG)

            elif self.xc == 'range_RPA':

                fv = np.exp(-0.25 * (G_G * self.range_rc) ** 2.0)

            elif self.linear_kernel:
                r = ulm.open('fhxc_%s_%s_%s_%s.ulm' %
                             (self.tag, self.xc, self.ecut_max, qi))
                fv = r.fhxc_sGsG

                if cut_G is not None:
                    fv = fv.take(cut_G, 0).take(cut_G, 1)

            elif not self.dyn_kernel:
                # static kernel which does not scale with lambda

                r = ulm.open('fhxc_%s_%s_%s_%s.ulm' %
                             (self.tag, self.xc, self.ecut_max, qi))
                fv = r.fhxc_lGG

                if cut_G is not None:
                    fv = fv.take(cut_G, 1).take(cut_G, 2)

            else:  # dynamical kernel
                r = ulm.open('fhxc_%s_%s_%s_%s.ulm' %
                             (self.tag, self.xc, self.ecut_max, qi))
                fv = r.fhxc_lwGG

                if cut_G is not None:
                    fv = fv.take(cut_G, 2).take(cut_G, 3)

            if pd.kd.gamma:
                G_G[0] = 1.0

            # Loop over frequencies; since the kernel has no spin,
            # we work with spin-summed response function
            e_w = []
            iw = 0

            for chi0_sGG in np.swapaxes(chi0_swGG, 0, 1):
                if cut_G is not None:
                    chi0_sGG = chi0_sGG.take(cut_G, 1).take(cut_G, 2)
                chi0v = np.zeros((nG, nG), dtype=complex)
                for s in range(ns):
                    chi0v += chi0_sGG[s] / G_G / G_G[:, np.newaxis]
                chi0v *= 4 * np.pi
                del chi0_sGG

                e = 0.0

                if not self.linear_kernel:

                    il = 0
                    for l, weight in zip(self.l_l, self.weight_l):

                        if not self.dyn_kernel:
                            chiv = np.linalg.solve(np.eye(nG) -
                                                   np.dot(chi0v, fv[il]),
                                                   chi0v).real
                        else:
                            chiv = np.linalg.solve(np.eye(nG) -
                                                   np.dot(chi0v, fv[il][iw]),
                                                   chi0v).real
                        e -= np.trace(chiv) * weight
                        il += 1

                    e += np.trace(chi0v.real)

                    e_w.append(e)

                    iw += 1

                else:

                    # Coupling constant integration
                    # for long-range part
                    # Do this analytically, except for the RPA
                    # simply since the analytical method is already
                    # implemented in rpa.py
                    if self.xc == 'range_RPA':
                        # way faster than np.dot for diagonal kernels
                        e_GG = np.eye(nG) - chi0v * fv
                    elif self.xc != 'RPA':
                        e_GG = np.eye(nG) - np.dot(chi0v, fv)

                    if self.xc == 'RPA':
                        # numerical RPA
                        elong = 0.0
                        for l, weight in zip(self.l_l, self.weight_l):

                            chiv = np.linalg.solve(np.eye(nG) - l *
                                                   np.dot(chi0v, fv),
                                                   chi0v).real

                            elong -= np.trace(chiv) * weight

                        elong += np.trace(chi0v.real)

                    else:
                        # analytic everything else
                        elong = (np.log(np.linalg.det(e_GG)) +
                                 nG - np.trace(e_GG)).real
                    # Numerical integration for short-range part
                    eshort = 0.0
                    if self.xc not in ('RPA', 'range_RPA', 'range_rALDA'):
                        fxcv = fv - np.eye(nG)  # Subtract Hartree contribution

                        for l, weight in zip(self.l_l, self.weight_l):

                            chiv = np.linalg.solve(np.eye(nG) - l *
                                                   np.dot(chi0v, fv), chi0v)
                            eshort += (np.trace(np.dot(chiv, fxcv)).real *
                                       weight)

                        eshort -= np.trace(np.dot(chi0v, fxcv)).real

                    elif self.xc in ('range_RPA', 'range_rALDA'):
                        eshort = (2 * np.pi * self.shortrange /
                                  np.sum(self.weight_w))

                    e = eshort + elong
                    e_w.append(e)

        E_w = np.zeros_like(self.omega_w)
        self.blockcomm.all_gather(np.array(e_w), E_w)
        energy = np.dot(E_w, self.weight_w) / (2 * np.pi)
        return energy

class KernelWave:

    def __init__(self, calc, xc, ibzq_qc, fd, l_l, q_empty,
                 omega_w, Eg, ecut, tag, timer):

        self.calc = calc
        self.gd = calc.density.gd
        self.xc = xc
        self.ibzq_qc = ibzq_qc
        self.fd = fd
        self.l_l = l_l
        self.ns = calc.wfs.nspins
        self.q_empty = q_empty
        self.omega_w = omega_w
        self.Eg = Eg
        self.ecut = ecut
        self.tag = tag
        self.timer = timer

        if l_l is None:  # -> Kernel is linear in coupling strength
            self.l_l = [1.0]

        # Density grid
        self.n_g = calc.get_all_electron_density(gridrefinement=2)
        self.n_g = self.n_g.flatten() * Bohr**3
        # Density now in units electrons/cubic Bohr
        #  For atoms with large vacuum regions
        #  this apparently can take negative values!
        mindens = np.amin(self.n_g)

        if mindens < 0:
            print('Negative densities found! (magnitude %s)'
                  % np.abs(mindens), file=self.fd)
            print('These will be reset to 1E-12 elec/bohr^3)', file=self.fd)
            self.n_g[np.where(self.n_g < 0.0)] = 1.0E-12

        r_g = self.gd.refine().get_grid_point_coordinates()  # already in Bohr
        self.x_g = 1.0 * r_g[0].flatten()
        self.y_g = 1.0 * r_g[1].flatten()
        self.z_g = 1.0 * r_g[2].flatten()
        self.gridsize = len(self.x_g)
        assert len(self.n_g) == self.gridsize

        if self.omega_w is not None:
            print('Calculating dynamical kernel at %s frequencies'
                  % len(self.omega_w), file=self.fd)

        if self.Eg is not None:
            print('Band gap of %s eV used to evaluate kernel'
                  % (self.Eg * Ha), file=self.fd)

        # Enhancement factor for GGA
        if self.xc == 'rAPBE' or self.xc == 'rAPBEns':
            nf_g = calc.get_all_electron_density(gridrefinement=4)
            nf_g *= Bohr**3
            gdf = self.gd.refine().refine()
            grad_v = [Gradient(gdf, v, n=1).apply for v in range(3)]
            gradnf_vg = gdf.empty(3)

            for v in range(3):
                grad_v[v](nf_g, gradnf_vg[v])

            self.s2_g = np.sqrt(np.sum(gradnf_vg[:, ::2, ::2, ::2]**2.0,
                                       0)).flatten()  # |\nabla\rho|
            self.s2_g *= 1.0 / (2.0 * (3.0 * np.pi**2.0)**(1.0 / 3.0)
                                * self.n_g**(4.0/3.0))
            # |\nabla\rho|/(2kF\rho) = s
            self.s2_g = self.s2_g**2  # s^2
            assert len(self.n_g) == len(self.s2_g)

            # Now we find all the regions where the
            # APBE kernel wants to be positive, and hack s to = 0,
            # so that we are really using the ALDA kernel
            # at these points
            apbe_g = self.get_PBE_fxc(self.n_g, self.s2_g)
            poskern_ind = np.where(apbe_g >= 0.0)
            if len(poskern_ind[0]) > 0:
                print('The APBE kernel takes positive values at ' +
                      '%s grid points out of a total of %s (%3.2f%%).' %
                      (len(poskern_ind[0]), self.gridsize,
                       100.0 * len(poskern_ind[0])/self.gridsize),
                      file=self.fd)
                print('The ALDA kernel will be used at these points',
                      file=self.fd)
                self.s2_g[poskern_ind] = 0.0

    def calculate_fhxc(self):

        print('Calculating %s kernel at %d eV cutoff' %
              (self.xc, self.ecut), file=self.fd)
        self.fd.flush()

        for iq, q_c in enumerate(self.ibzq_qc):

            if iq < self.q_empty:  # don't recalculate q vectors
                continue

            thisqd = KPointDescriptor([q_c])
            pd = PWDescriptor(self.ecut/Ha, self.gd, complex, thisqd)

            nG = pd.ngmax
            G_G = pd.G2_qG[0]**0.5  # |G+q|
            Gv_G = pd.get_reciprocal_vectors(q=0, add_q=False)
            # G as a vector (note we are at a specific q point here so set q=0)

            # Distribute G vectors among processors
            # Later we calculate for iG' > iG,
            # so stagger allocation in order to balance load
            local_Gvec_grid_size = nG//mpi.world.size
            my_Gints = (mpi.world.rank
                        + np.arange(0,
                                    local_Gvec_grid_size*mpi.world.size,
                                    mpi.world.size))

            if (mpi.world.rank + (local_Gvec_grid_size)*mpi.world.size) < nG:
                my_Gints = np.append(my_Gints, [mpi.world.rank +
                                                local_Gvec_grid_size *
                                                mpi.world.size])

            my_Gv_G = Gv_G[my_Gints]

            if (self.ns == 2) and (self.xc == 'rALDA' or self.xc == 'rAPBE'):

                assert len(self.l_l) == 1

                # Form spin-dependent kernel according to
                # PRB 88, 115131 (2013) equation 20
                # (note typo, should be \tilde{f^rALDA})
                # spincorr is just the ALDA exchange kernel
                # with a step function (\equiv \tilde{f^rALDA})
                # fHxc^{up up}     = fHxc^{down down} = fv_nospin + fv_spincorr
                # fHxc^{up down}   = fHxc^{down up}   = fv_nospin - fv_spincorr

                calc_spincorr = True
                fv_spincorr = np.zeros((nG, nG), dtype=complex)

            else:

                calc_spincorr = False

            if self.omega_w is None:
                fv_nospin = np.zeros((len(self.l_l), nG, nG), dtype=complex)
            else:
                fv_nospin = np.zeros((len(self.l_l),
                                     len(self.omega_w), nG, nG), dtype=complex)

            for il, l in enumerate(self.l_l):  # loop over coupling constant

                for iG, Gv in zip(my_Gints, my_Gv_G):  # loop over G vecs

                    # For all kernels except JGM we
                    # treat head and wings analytically
                    if G_G[iG] > 1.0E-5 or self.xc == 'JGMs':

                        # Symmetrised |q+G||q+G'|, where iG' >= iG
                        mod_Gpq = np.sqrt(G_G[iG]*G_G[iG:])

                        # Phase factor \vec{G}-\vec{G'}
                        deltaGv = Gv - Gv_G[iG:]

                        if (self.xc in ('rALDA', 'range_rALDA', 'rALDAns')):

                            # rALDA trick: the Hartree-XC kernel is exactly
                            # zero for densities below rho_min =
                            # min_Gpq^3/(24*pi^2),
                            # so we don't need to include these contributions
                            # in the Fourier transform

                            min_Gpq = np.amin(mod_Gpq)
                            rho_min = min_Gpq**3.0/(24.0 * np.pi**2.0)
                            small_ind = np.where(self.n_g >= rho_min)

                        elif (self.xc == 'rAPBE' or self.xc == 'rAPBEns'):

                            # rAPBE trick: the Hartree-XC kernel
                            # is exactly zero at grid points where
                            # min_Gpq > cutoff wavevector

                            min_Gpq = np.amin(mod_Gpq)
                            small_ind = np.where(min_Gpq <=
                                                 np.sqrt(-4.0 * np.pi /
                                                         self.get_PBE_fxc(self.n_g, self.s2_g)))

                        else:

                            small_ind = np.arange(self.gridsize)

                        phase_Gpq = np.exp(-1.0j*(deltaGv[:, 0, np.newaxis] *
                                                  self.x_g[small_ind]
                                                  + deltaGv[:, 1, np.newaxis] *
                                                  self.y_g[small_ind]
                                                  + deltaGv[:, 2, np.newaxis] *
                                                  self.z_g[small_ind]))
                        if self.omega_w is None:
                            fv_nospin[il, iG, iG:] = self.get_scaled_fHxc_q(
                                q=mod_Gpq,
                                sel_points=small_ind,
                                Gphase=phase_Gpq,
                                l=l,
                                spincorr=False,
                                w=None)
                        else:
                            for iw, omega in enumerate(self.omega_w):
                                fv_nospin[il, iw, iG, iG:] = \
                                    self.get_scaled_fHxc_q(
                                        q=mod_Gpq,
                                        sel_points=small_ind,
                                        Gphase=phase_Gpq,
                                        l=l,
                                        spincorr=False,
                                        w=omega)

                        if calc_spincorr:
                            fv_spincorr[iG, iG:] = self.get_scaled_fHxc_q(
                                q=mod_Gpq,
                                sel_points=small_ind,
                                Gphase=phase_Gpq,
                                l=1.0,
                                spincorr=True,
                                w=None)

                    else:
                        # head and wings of q=0 are dominated by
                        # 1/q^2 divergence of scaled Coulomb interaction

                        assert iG == 0

                        if self.omega_w is None:
                            fv_nospin[il, 0, 0] = l
                            fv_nospin[il, 0, 1:] = 0.0
                        else:
                            fv_nospin[il, :, 0, 0] = l
                            fv_nospin[il, :, 0, 1:] = 0.0

                        if calc_spincorr:
                            fv_spincorr[0, :] = 0.0

                    # End loop over G vectors

                mpi.world.sum(fv_nospin[il])

                if self.omega_w is None:
                    # We've only got half the matrix here,
                    # so add the hermitian conjugate:
                    fv_nospin[il] += np.conj(fv_nospin[il].T)
                    # but now the diagonal's been doubled,
                    # so we multiply these elements by 0.5
                    fv_nospin[il][np.diag_indices(nG)] *= 0.5

                else:  # same procedure for dynamical kernels
                    for iw in range(len(self.omega_w)):
                        fv_nospin[il][iw] += np.conj(fv_nospin[il][iw].T)
                        fv_nospin[il][iw][np.diag_indices(nG)] *= 0.5

                # End of loop over coupling constant

            if calc_spincorr:
                mpi.world.sum(fv_spincorr)
                fv_spincorr += np.conj(fv_spincorr.T)
                fv_spincorr[np.diag_indices(nG)] *= 0.5

            # Write to disk
            if mpi.rank == 0:

                w = ulm.open('fhxc_%s_%s_%s_%s.ulm' %
                             (self.tag, self.xc, self.ecut, iq), 'w')

                if calc_spincorr:
                    # Form the block matrix kernel
                    fv_full = np.empty((2 * nG, 2 * nG), dtype=complex)
                    fv_full[:nG, :nG] = fv_nospin[0] + fv_spincorr
                    fv_full[:nG, nG:] = fv_nospin[0] - fv_spincorr
                    fv_full[nG:, :nG] = fv_nospin[0] - fv_spincorr
                    fv_full[nG:, nG:] = fv_nospin[0] + fv_spincorr
                    w.write(fhxc_sGsG=fv_full)

                elif len(self.l_l) == 1:
                    w.write(fhxc_sGsG=fv_nospin[0])

                elif self.omega_w is None:
                    w.write(fhxc_lGG=fv_nospin)

                else:
                    w.write(fhxc_lwGG=fv_nospin)
                w.close()

            print('q point %s complete' % iq, file=self.fd)
            self.fd.flush()

            mpi.world.barrier()

    def get_scaled_fHxc_q(self, q, sel_points, Gphase, l, spincorr, w):
        # Given a coupling constant l, construct the Hartree-XC
        # kernel in q space a la Lein, Gross and Perdew,
        # Phys. Rev. B 61, 13431 (2000):
        #
        # f_{Hxc}^\lambda(q,\omega,r_s) = \frac{4\pi \lambda }{q^2}  +
        # \frac{1}{\lambda} f_{xc}(q/\lambda,\omega/\lambda^2,\lambda r_s)
        #
        # divided by the unscaled Coulomb interaction!!
        #
        # i.e. this subroutine returns f_{Hxc}^\lambda(q,\omega,r_s)
        #                              *  \frac{q^2}{4\pi}
        # = \lambda * [\frac{(q/lambda)^2}{4\pi}
        #              f_{Hxc}(q/\lambda,\omega/\lambda^2,\lambda r_s)]
        # = \lambda * [1/scaled_coulomb * fHxc computed with scaled quantities]

        # Apply scaling
        rho = self.n_g[sel_points]

        # GGA enhancement factor s is lambda independent,
        # but we might want to truncate it
        if self.xc == 'rAPBE' or self.xc == 'rAPBEns':
            s2_g = self.s2_g[sel_points]
        else:
            s2_g = None

        scaled_q = q / l
        scaled_rho = rho / l**3.0
        scaled_rs = (3.0/(4.0*np.pi*scaled_rho))**(1.0/3.0)  # Wigner radius
        if w is not None:
            scaled_w = w/(l**2.0)
        else:
            scaled_w = None

        if self.Eg is not None:
            scaled_Eg = self.Eg/(l**1.5)
        else:
            scaled_Eg = None

        if not spincorr:
            scaled_kernel = l * self.get_fHxc_q(scaled_rs, scaled_q,
                                                Gphase, s2_g,
                                                scaled_w, scaled_Eg)
        else:
            scaled_kernel = l * self.get_spinfHxc_q(scaled_rs, scaled_q,
                                                    Gphase, s2_g)

        return(scaled_kernel)

    def get_fHxc_q(self, rs, q, Gphase, s2_g, w, scaled_Eg):
        # Construct fHxc(q,G,:), divided by scaled Coulomb interaction

        qF = (9.0*np.pi/4.0)**(1.0/3.0) * 1.0/rs

        if self.xc in ('rALDA', 'rALDAns', 'range_rALDA'):
            # rALDA (exchange only) kernel
            # Olsen and Thygesen, Phys. Rev. B 88, 115131 (2013)
            # ALDA up to 2*qF, -vc for q >2qF (such that fHxc vanishes)

            rxalda_A = 0.25
            rxalda_qcut = qF * np.sqrt(1.0/rxalda_A)

            # construct fHxc(k,r)
            fHxc_Gr = (0.5 + 0.0j) * ((1.0 + np.sign(rxalda_qcut -
                                                     q[:, np.newaxis])) *
                                      (1.0 + (-1.0) * rxalda_A *
                                       (q[:, np.newaxis]/qF)**2.0))

        elif self.xc == 'rALDAc':
            # rALDA exchange+correlation kernel
            # Olsen and Thygesen, Phys. Rev. B 88, 115131 (2013)
            # Cutoff wavevector ensures kernel is continuous

            rxcalda_A = self.get_heg_A(rs)
            rxcalda_qcut = qF * np.sqrt(1.0/rxcalda_A)

            # construct fHxc(k,r)
            fHxc_Gr = (0.5 + 0.0j) * ((1.0 + np.sign(rxcalda_qcut -
                                                     q[:, np.newaxis])) *
                                      (1.0 + (-1.0) * rxcalda_A *
                                       (q[:, np.newaxis]/qF)**2.0))

        elif self.xc == 'rAPBE' or self.xc == 'rAPBEns':

            # Olsen and Thygesen, Phys. Rev. Lett. 112, 203001 (2014)
            # Exchange only part of the PBE XC kernel, neglecting the terms
            # arising from the variation of the density gradient
            # i.e. second functional derivative
            # d2/drho^2 -> \partial^2/\partial rho^2 at fixed \nabla \rho

            fxc_PBE = self.get_PBE_fxc(pbe_rho=3.0/(4.0 * np.pi * rs**3.0),
                                       pbe_s2_g=s2_g)
            rxapbe_qcut = np.sqrt(-4.0 * np.pi / fxc_PBE)

            fHxc_Gr = (0.5 + 0.0j) * ((1.0 + np.sign(rxapbe_qcut -
                                                     q[:, np.newaxis])) *
                                      (1.0 + fxc_PBE / (4.0*np.pi) *
                                       (q[:, np.newaxis])**2.0))

        elif self.xc == 'CP':
            # Constantin & Pitarke, Phys. Rev. B 75, 245127 (2007)
            # equation 4, omega = 0
            # and equation 9 [note that fxc(q=0) = fxc^ALDA = -4piA/qf^2]
            # The simplest, static error-function kernel.
            # Produces correct q = 0 limit, but not q->oo

            cp_kappa = self.get_heg_A(rs)/(qF**2.0)
            fHxc_Gr = (1.0 + 0.0j) * np.exp(-cp_kappa * q[:, np.newaxis]**2.0)

        elif self.xc == 'CP_dyn':
            # CP kernel with frequency dependence

            cp_A = self.get_heg_A(rs)
            cp_D = self.get_heg_D(rs)
            cp_c = cp_D/cp_A
            cp_a = 6.0*np.sqrt(cp_c)

            cp_kappa_w = cp_A/(qF**2.0)
            cp_kappa_w *= (1.0 + cp_a*w + cp_c*w**2.0)/(1.0 + w**2.0)

            fHxc_Gr = (1.0 + 0.0j) * np.exp(-cp_kappa_w *
                                            q[:, np.newaxis]**2.0)

        elif self.xc == 'CDOP' or self.xc == 'CDOPs':
            # Corradini, Del Sole, Onida and Palummo (CDOP),
            # Phys. Rev. B 57, 14569 (1998)
            # Reproduces exact q=0 and q -> oo limits

            cdop_Q = q[:, np.newaxis]/qF
            cdop_A = self.get_heg_A(rs)
            cdop_B = self.get_heg_B(rs)
            if self.xc == 'CDOPs':
                cdop_C = np.zeros(np.shape(cdop_B))
            else:
                cdop_C = self.get_heg_C(rs)
            cdop_g = cdop_B/(cdop_A - cdop_C)
            cdop_alpha = 1.5/(rs**0.25) * cdop_A/(cdop_B*cdop_g)
            cdop_beta = 1.2/(cdop_B * cdop_g)

            fHxc_Gr = -cdop_C * cdop_Q**2.0 * (1.0 + 0.0j)
            fHxc_Gr += -cdop_B * cdop_Q**2.0 * 1.0/(cdop_g + cdop_Q**2.0)
            fHxc_Gr += -(cdop_alpha * cdop_Q**4.0 *
                         np.exp(-1.0*cdop_beta*cdop_Q**2.0))

            # Hartree
            fHxc_Gr += 1.0

        elif self.xc == 'JGMs':
            # Trevisanutto et al.,
            # Phys. Rev. B 87, 205143 (2013)
            # equation 4,
            # so-called "jellium with gap" model, but simplified
            # so that it is similar to CP.

            jgm_kappa = self.get_heg_A(rs)/(qF**2.0)
            jgm_n = 3.0/(4.0*np.pi * rs**3.0)
            fHxc_Gr = (1.0 + 0.0j) * (np.exp(-jgm_kappa *
                                      q[:, np.newaxis]**2.0) *
                                      np.exp(-scaled_Eg**2.0/(4.0*np.pi*jgm_n)))

        # Integrate over r with phase
        fHxc_Gr *= Gphase
        fHxc_GG = np.sum(fHxc_Gr, 1) / self.gridsize
        return(fHxc_GG)

    def get_spinfHxc_q(self, rs, q, Gphase, s2_g):

        qF = (9.0 * np.pi / 4.0)**(1.0 / 3.0) * 1.0 / rs

        if self.xc == 'rALDA':

            rxalda_A = 0.25
            rxalda_qcut = qF * np.sqrt(1.0/rxalda_A)

            fspinHxc_Gr = ((0.5 + 0.0j) *
                           (1.0 + np.sign(rxalda_qcut - q[:, np.newaxis])) *
                           (-1.0) * rxalda_A * (q[:, np.newaxis]/qF)**2.0)

        elif self.xc == 'rAPBE':

            fxc_PBE = self.get_PBE_fxc(pbe_rho=(3.0 /
                                       (4.0 * np.pi * rs**3.0)), pbe_s2_g=s2_g)
            rxapbe_qcut = np.sqrt(-4.0 * np.pi / fxc_PBE)

            fspinHxc_Gr = ((0.5 + 0.0j) *
                           (1.0 + np.sign(rxapbe_qcut - q[:, np.newaxis])) *
                           fxc_PBE / (4.0 * np.pi) *
                           q[:, np.newaxis]**2.0)

        fspinHxc_Gr *= Gphase
        fspinHxc_GG = np.sum(fspinHxc_Gr, 1) / self.gridsize
        return(fspinHxc_GG)

    def get_PBE_fxc(self, pbe_rho, pbe_s2_g):

        pbe_kappa = 0.804
        pbe_mu = 0.2195149727645171

        pbe_denom_g = 1.0 + pbe_mu * pbe_s2_g / pbe_kappa

        F_g = 1.0 + pbe_kappa - pbe_kappa / pbe_denom_g
        Fn_g = -8.0 / 3.0 * pbe_mu * pbe_s2_g / pbe_rho / pbe_denom_g**2.0
        Fnn_g = (-11.0 / 3.0 / pbe_rho * Fn_g -
                 2.0 / pbe_kappa * Fn_g**2.0 * pbe_denom_g)

        e_g = -3.0 / 4.0 * (3.0 / np.pi)**(1.0 / 3.0) * pbe_rho**(4.0/3.0)
        v_g = 4.0 / 3.0 * e_g / pbe_rho
        f_g = 1.0 / 3.0 * v_g / pbe_rho

        pbe_f_g = f_g * F_g + 2.0 * v_g * Fn_g + e_g * Fnn_g

        return(pbe_f_g)

    def get_heg_A(self, rs):
        # Returns the A coefficient, where the
        # q ->0 limiting value of static fxc
        # of the HEG = -\frac{4\pi A }{q_F^2} = f_{xc}^{ALDA}.
        # We need correlation energy per electron and first and second derivs
        # w.r.t. rs
        # See for instance Moroni, Ceperley and Senatore,
        # Phys. Rev. Lett. 75, 689 (1995)
        # (and also Kohn and Sham, Phys. Rev. 140, A1133 (1965) equation 2.7)

        # Exchange contribution
        heg_A = 0.25

        # Correlation contribution
        A_ec, A_dec, A_d2ec = self.get_pw_lda(rs)
        heg_A += (1.0/27.0 * rs**2.0 * (9.0 * np.pi / 4.0)**(2.0/3.0)
                  * (2 * A_dec - rs*A_d2ec))

        return(heg_A)

    def get_heg_B(self, rs):
        # Returns the B coefficient, where the
        # q -> oo limiting behaviour of static fxc
        # of the HEG is -\frac{4\pi B}{q^2} - \frac{4\pi C}{q_F^2}.
        # Use the parametrisation of Moroni, Ceperley and Senatore,
        # Phys. Rev. Lett. 75, 689 (1995)

        mcs_xs = np.sqrt(rs)

        mcs_a = (1.0, 2.15, 0.0, 0.435)
        mcs_b = (3.0, 1.57, 0.0, 0.409)

        mcs_num = 0

        for mcs_j, mcs_coeff in enumerate(mcs_a):
            mcs_num += mcs_coeff * mcs_xs**mcs_j

        mcs_denom = 0

        for mcs_j, mcs_coeff in enumerate(mcs_b):
            mcs_denom += mcs_coeff * mcs_xs**mcs_j

        heg_B = mcs_num / mcs_denom
        return(heg_B)

    def get_heg_C(self, rs):
        # Returns the C coefficient, where the
        # q -> oo limiting behaviour of static fxc
        # of the HEG is -\frac{4\pi B}{q^2} - \frac{4\pi C}{q_F^2}.
        # Again see Moroni, Ceperley and Senatore,
        # Phys. Rev. Lett. 75, 689 (1995)

        C_ec, C_dec, Cd2ec = self.get_pw_lda(rs)

        heg_C = ((-1.0) * np.pi**(2.0 / 3.0)*(1.0 / 18.0)**(1.0 / 3.0) *
                 (rs*C_ec + rs**2.0 * C_dec))

        return(heg_C)

    def get_heg_D(self, rs):
        # Returns a 'D' coefficient, where the
        # q->0 omega -> oo limiting behaviour
        # of the frequency dependent fxc is -\frac{4\pi D}{q_F^2}
        # see Constantin & Pitarke Phys. Rev. B 75, 245127 (2007) equation 7

        D_ec, D_dec, D_d2ec = self.get_pw_lda(rs)

        # Exchange contribution
        heg_D = 0.15
        # Correlation contribution
        heg_D += ((9.0 * np.pi / 4.0)**(2.0 / 3.0) * rs / 3.0 *
                  (22.0 / 15.0 * D_ec + 26.0 / 15.0 * rs * D_dec))
        return(heg_D)

    def get_pw_lda(self, rs):
        # Returns LDA correlation energy and its first and second
        # derivatives with respect to rs according to the parametrisation
        # of Perdew and Wang, Phys. Rev. B 45, 13244 (1992)

        pw_A = 0.031091
        pw_alp = 0.21370
        pw_beta = (7.5957, 3.5876, 1.6382, 0.49294)

        pw_logdenom = 2.0 * pw_A * (pw_beta[0]*rs**0.5 +
                                    pw_beta[1]*rs**1.0 +
                                    pw_beta[2]*rs**1.5 +
                                    pw_beta[3]*rs**2.0)

        pw_dlogdenom = 2.0 * pw_A * (0.5*pw_beta[0]*rs**(-0.5) +
                                     1.0*pw_beta[1] +
                                     1.5*pw_beta[2] * rs**(0.5) +
                                     2.0*pw_beta[3] * rs)

        pw_d2logdenom = 2.0 * pw_A * (-0.25 * pw_beta[0] * rs**(-1.5) +
                                      0.75 * pw_beta[2] * rs**(-0.5) +
                                      2.0 * pw_beta[3])

        pw_logarg = 1.0 + 1.0 / pw_logdenom
        pw_dlogarg = (-1.0) / (pw_logdenom**2.0) * pw_dlogdenom
        pw_d2logarg = 2.0 / (pw_logdenom**3.0) * (pw_dlogdenom**2.0)
        pw_d2logarg += (-1.0) / (pw_logdenom**2.0) * pw_d2logdenom

        # pw_ec = the correlation energy (per electron)
        pw_ec = -2.0 * pw_A * (1 + pw_alp * rs) * np.log(pw_logarg)

        # pw_dec = first derivative

        pw_dec = -2.0 * pw_A * (1 + pw_alp * rs) * pw_dlogarg / pw_logarg
        pw_dec += (-2.0 * pw_A * pw_alp) * np.log(pw_logarg)

        # pw_d2ec = second derivative

        pw_d2ec = (-2.0) * pw_A * pw_alp * pw_dlogarg / pw_logarg
        pw_d2ec += (-2.0) * pw_A * ((1 + pw_alp * rs) *
                                    (pw_d2logarg / pw_logarg -
                                     (pw_dlogarg**2.0) / (pw_logarg**2.0)))
        pw_d2ec += (-2.0 * pw_A * pw_alp) * pw_dlogarg / pw_logarg

        return((pw_ec, pw_dec, pw_d2ec))


class range_separated:
    def __init__(self, calc, fd, frequencies, freqweights, l_l, lweights, range_rc, xc):

        self.calc = calc
        self.fd = fd
        self.frequencies = frequencies
        self.freqweights = freqweights
        self.l_l = l_l
        self.lweights = lweights
        self.range_rc = range_rc
        self.xc = xc

        self.cutoff_rs = 36.278317

        if self.xc == 'range_RPA':
            print('Using range-separated RPA approach, with parameter %s Bohr' %
                  (self.range_rc), file=self.fd)

        nval_g = calc.get_all_electron_density(gridrefinement=4, skip_core=True
                                               ).flatten() * Bohr**3
        self.dv = self.calc.density.gd.dv / 64.0  # 64 = gridrefinement^3

        density_cut = 3.0 / (4.0 * np.pi * self.cutoff_rs**3.0)
        if (nval_g < 0.0).any():
            print('Warning, negative densities found! (Magnitude %s)' %
                  np.abs(np.amin(nval_g)), file=self.fd)
            print('These will be ignored', file=self.fd)
        if (nval_g < density_cut).any():
            nval_g = nval_g[np.where(nval_g > density_cut)]
            print('Not calculating correlation energy ',
                  'contribution for densities < %3.2e elecs/Bohr ^ 3' %
                  (density_cut), file=self.fd)

        densitysum = np.sum(nval_g*self.dv)
        valence = self.calc.wfs.setups.nvalence

        print('Density integrates to %s electrons' %
              (densitysum), file=self.fd)

        print('Renormalized to %s electrons' %
              (valence), file=self.fd)

        nval_g *= valence / densitysum
        self.rs_g = (3.0 / (4.0 * np.pi * nval_g))**(1.0 / 3.0)

        self.rsmin = np.amin(self.rs_g)
        self.rsmax = np.amax(self.rs_g)

    def calculate(self):

        print('Generating tables of electron gas energies...', file=self.fd)

        table_SR = self.generate_tables()

        print('...done', file=self.fd)
        # Now interpolate the table to calculate local density terms
        E_SR = np.sum(np.interp(self.rs_g, table_SR[:, 0],
                                table_SR[:, 1])) * self.dv

        # RPA energy minus long range correlation
        print('Short range correlation energy/unit cell = %5.4f eV \n' %
              ((E_SR) * Ha), file=self.fd)
        e = E_SR

        return(e)

    def generate_tables(self):

        # Finite difference steps for density and k vec
        rs_step = 0.01
        k_step = 0.01

        rs_r = np.arange(self.rsmin - rs_step, self.rsmax + rs_step,
                         rs_step)

        table_SR = np.empty((len(rs_r), 2))
        table_SR[:, 0] = rs_r
        for iR, Rs in enumerate(rs_r):

            qF = (9.0 * np.pi / 4.0)**(1.0 / 3.0) * 1.0 / Rs

            q_k = np.arange(k_step, 10.0 * qF, k_step)

            if self.xc == 'range_RPA':
                # Correlation energy per electron, in Hartree, per k
                Eeff_k, Erpa_k = self.RPA_corr_hole(q_k, Rs)
                ESR_k = Erpa_k - Eeff_k
            elif self.xc == 'range_rALDA':
                ESR_k = self.rALDA_corr_hole(q_k, Rs)

            # Integrate over k
            table_SR[iR, 1] = k_step * np.sum(ESR_k)

        return table_SR

    def RPA_corr_hole(self, q, rs):

        # Integrating this quantity over q, gives
        # correlation energy per unit volume
        # calcuated with a Coulomb-like effective
        # interaction, in Hartree
        # = 1/(2\pi) * \sum_{\vec{q}} \int_0^infty ds
        #   * [ ln (1 - v_eff \chi_0) + v_eff \chi_0]
        # = 1/(2\pi) * \int 4 \pi q^2 dq /((2\pi)^3)
        #   * \int_0^infty ds [ ln (1 - v_eff \chi_0) + v_eff \chi_0]
        # = 1/(4\pi^3) * \int q^2 dq  \int_0^infty ds
        #   * [ ln (1 - v_eff \chi_0) + v_eff \chi_0]

        veff = 4.0 * np.pi / (q * q) * np.exp(-0.25 * q * q *
                                              self.range_rc * self.range_rc)
        vc = 4.0 * np.pi / (q * q)

        # Do the integral over frequency using Gauss-Legendre

        eeff_q = np.zeros(len(q))
        erpa_q = np.zeros(len(q))

        for u, freqweight in zip(self.frequencies, self.freqweights):

            chi0 = self.calc_lindhard(q, u, rs)

            eff_integrand = np.log(np.ones(len(q)) - veff * chi0) + veff * chi0
            eeff_q += eff_integrand * freqweight

            rpa_integrand = np.log(np.ones(len(q)) - vc * chi0) + vc * chi0
            erpa_q += rpa_integrand * freqweight

        # Per unit volume

        eeff_q *= 1.0/(4.0 * np.pi**3.0) * q * q
        erpa_q *= 1.0/(4.0 * np.pi**3.0) * q * q

        return(eeff_q, erpa_q)

    def rALDA_corr_hole(self, q, rs):
        qF = (9.0 * np.pi / 4.0)**(1.0 / 3.0) * 1.0 / rs

        veff = 4.0 * np.pi / (q * q) * ((1.0 - 0.25 * q * q / (qF * qF)) *
                                        0.5 * (1.0 + np.sign(2.0 * qF - q)))
        fxc = veff - 4.0 * np.pi / (q * q)

        esr_q = np.zeros(len(q))

        for u, freqweight in zip(self.frequencies, self.freqweights):

            chi0 = self.calc_lindhard(q, u, rs)

            esr_u = np.zeros(len(q))

            for l, lweight in zip(self.l_l, self.lweights):

                chil = chi0 / (1.0 - l * fxc * chi0)
                esr_u += lweight * (-fxc) * (chil - chi0)

            esr_q += freqweight * esr_u

        esr_q *= 1.0/(4.0 * np.pi**3.0) * q * q

        return(esr_q)

    def calc_lindhard(self, q, u, rs):

        # Calculate Lindhard function at imaginary frequency u

        qF = (9.0 * np.pi / 4.0)**(1.0 / 3.0) * 1.0 / rs

        Q = q / 2.0 / qF
        U = u / q / qF
        lchi = ((Q * Q - U * U - 1.0) / 4.0 / Q * np.log((U * U +
                (Q + 1.0) * (Q + 1.0))/(U * U + (Q - 1.0) * (Q - 1.0))))
        lchi += -1.0 + U * np.arctan((1.0 + Q)/U) + U*np.arctan((1.0 - Q)/U)
        lchi *= qF/(2.0 * np.pi * np.pi)
        return(lchi)


class KernelDens:
    def __init__(self, calc, xc, ibzq_qc, fd, unit_cells,
                 density_cut, ecut, tag, timer):

        self.calc = calc
        self.gd = calc.density.gd
        self.xc = xc
        self.ibzq_qc = ibzq_qc
        self.fd = fd
        self.unit_cells = unit_cells
        self.density_cut = density_cut
        self.ecut = ecut
        self.tag = tag
        self.timer = timer

        self.A_x = -(3 / 4.) * (3 / np.pi)**(1 / 3.)

        self.n_g = calc.get_all_electron_density(gridrefinement=1)
        self.n_g *= Bohr**3

        if xc[-3:] == 'PBE':
            nf_g = calc.get_all_electron_density(gridrefinement=2)
            nf_g *= Bohr**3
            gdf = self.gd.refine()
            grad_v = [Gradient(gdf, v, n=1).apply for v in range(3)]
            gradnf_vg = gdf.empty(3)
            for v in range(3):
                grad_v[v](nf_g, gradnf_vg[v])
            self.gradn_vg = gradnf_vg[:, ::2, ::2, ::2]

        qd = KPointDescriptor(self.ibzq_qc)
        self.pd = PWDescriptor(ecut / Ha, self.gd, complex, qd)

    @timer('FHXC')
    def calculate_fhxc(self):

        print('Calculating %s kernel at %d eV cutoff' %
              (self.xc, self.ecut), file=self.fd)
        if self.xc[0] == 'r':
            self.calculate_rkernel()
        else:
            assert self.xc[0] == 'A'
            self.calculate_local_kernel()

    def calculate_rkernel(self):

        gd = self.gd
        ng_c = gd.N_c
        cell_cv = gd.cell_cv
        icell_cv = 2 * np.pi * np.linalg.inv(cell_cv)
        vol = np.linalg.det(cell_cv)

        ns = self.calc.wfs.nspins
        n_g = self.n_g   # density on rough grid

        fx_g = ns * self.get_fxc_g(n_g)   # local exchange kernel
        qc_g = (-4 * np.pi * ns / fx_g)**0.5   # cutoff functional
        flocal_g = qc_g**3 * fx_g / (6 * np.pi**2)   # ren. x-kernel for r=r'
        Vlocal_g = 2 * qc_g / np.pi   # ren. Hartree kernel for r=r'

        ng = np.prod(ng_c)   # number of grid points
        r_vg = gd.get_grid_point_coordinates()
        rx_g = r_vg[0].flatten()
        ry_g = r_vg[1].flatten()
        rz_g = r_vg[2].flatten()

        print('    %d grid points and %d plane waves at the Gamma point' %
              (ng, self.pd.ngmax), file=self.fd)

        # Unit cells
        R_Rv = []
        weight_R = []
        nR_v = self.unit_cells
        nR = np.prod(nR_v)
        for i in range(-nR_v[0] + 1, nR_v[0]):
            for j in range(-nR_v[1] + 1, nR_v[1]):
                for h in range(-nR_v[2] + 1, nR_v[2]):
                    R_Rv.append(i * cell_cv[0] +
                                j * cell_cv[1] +
                                h * cell_cv[2])
                    weight_R.append((nR_v[0] - abs(i)) *
                                    (nR_v[1] - abs(j)) *
                                    (nR_v[2] - abs(h)) / float(nR))
        if nR > 1:
            # with more than one unit cell only the exchange kernel is
            # calculated on the grid. The bare Coulomb kernel is added
            # in PW basis and Vlocal_g only the exchange part
            dv = self.calc.density.gd.dv
            gc = (3 * dv / 4 / np.pi)**(1 / 3.)
            Vlocal_g -= 2 * np.pi * gc**2 / dv
            print('    Lattice point sampling: ' +
                  '(%s x %s x %s)^2 ' % (nR_v[0], nR_v[1], nR_v[2]) +
                  ' Reduced to %s lattice points' % len(R_Rv), file=self.fd)

        l_g_size = -(-ng // mpi.world.size)
        l_g_range = range(mpi.world.rank * l_g_size,
                          min((mpi.world.rank + 1) * l_g_size, ng))

        fhxc_qsGr = {}
        for iq in range(len(self.ibzq_qc)):
            fhxc_qsGr[iq] = np.zeros((ns, len(self.pd.G2_qG[iq]),
                                      len(l_g_range)), dtype=complex)

        inv_error = np.seterr()
        np.seterr(invalid='ignore')
        np.seterr(divide='ignore')

        t0 = time()
        # Loop over Lattice points
        for i, R_v in enumerate(R_Rv):
            # Loop over r'. f_rr and V_rr are functions of r (dim. as r_vg[0])
            if i == 1:
                print('      Finished 1 cell in %s seconds' % int(time() - t0) +
                      ' - estimated %s seconds left' %
                      int((len(R_Rv) - 1) * (time() - t0)),
                      file=self.fd)
                self.fd.flush()
            if len(R_Rv) > 5:
                if (i + 1) % (len(R_Rv) / 5 + 1) == 0:
                    print('      Finished %s cells in %s seconds'
                          % (i, int(time() - t0))
                          + ' - estimated %s seconds left'
                          % int((len(R_Rv) - i) * (time() - t0) / i),
                          file=self.fd)
                    self.fd.flush()
            for g in l_g_range:
                rx = rx_g[g] + R_v[0]
                ry = ry_g[g] + R_v[1]
                rz = rz_g[g] + R_v[2]

                # |r-r'-R_i|
                rr = ((r_vg[0] - rx)**2 +
                      (r_vg[1] - ry)**2 +
                      (r_vg[2] - rz)**2)**0.5

                n_av = (n_g + n_g.flatten()[g]) / 2.
                fx_g = ns * self.get_fxc_g(n_av, index=g)
                qc_g = (-4 * np.pi * ns / fx_g)**0.5
                x = qc_g * rr
                osc_x = np.sin(x) - x * np.cos(x)
                f_rr = fx_g * osc_x / (2 * np.pi**2 * rr**3)
                if nR > 1:   # include only exchange part of the kernel here
                    V_rr = (sici(x)[0] * 2 / np.pi - 1) / rr
                else:        # include the full kernel (also hartree part)
                    V_rr = (sici(x)[0] * 2 / np.pi) / rr

                # Terms with r = r'
                if (np.abs(R_v) < 0.001).all():
                    tmp_flat = f_rr.flatten()
                    tmp_flat[g] = flocal_g.flatten()[g]
                    f_rr = tmp_flat.reshape(ng_c)
                    tmp_flat = V_rr.flatten()
                    tmp_flat[g] = Vlocal_g.flatten()[g]
                    V_rr = tmp_flat.reshape(ng_c)
                    del tmp_flat

                f_rr[np.where(n_av < self.density_cut)] = 0.0
                V_rr[np.where(n_av < self.density_cut)] = 0.0

                f_rr *= weight_R[i]
                V_rr *= weight_R[i]

                # r-r'-R_i
                r_r = np.array([r_vg[0] - rx, r_vg[1] - ry, r_vg[2] - rz])

                # Fourier transform of r
                for iq, q in enumerate(self.ibzq_qc):
                    q_v = np.dot(q, icell_cv)
                    e_q = np.exp(-1j * gemmdot(q_v, r_r, beta=0.0))
                    f_q = self.pd.fft((f_rr + V_rr) * e_q, iq) * vol / ng
                    fhxc_qsGr[iq][0, :, g - l_g_range[0]] += f_q
                    if ns == 2:
                        f_q = self.pd.fft(V_rr * e_q, iq) * vol / ng
                        fhxc_qsGr[iq][1, :, g - l_g_range[0]] += f_q

        mpi.world.barrier()

        np.seterr(**inv_error)

        for iq, q in enumerate(self.ibzq_qc):
            npw = len(self.pd.G2_qG[iq])
            fhxc_sGsG = np.zeros((ns * npw, ns * npw), complex)
            l_pw_size = -(-npw // mpi.world.size)  # parallelize over PW below
            l_pw_range = range(mpi.world.rank * l_pw_size,
                               min((mpi.world.rank + 1) * l_pw_size, npw))

            if mpi.world.size > 1:
                # redistribute grid and plane waves in fhxc_qsGr[iq]
                bg1 = BlacsGrid(mpi.world, 1, mpi.world.size)
                bg2 = BlacsGrid(mpi.world, mpi.world.size, 1)
                bd1 = bg1.new_descriptor(npw, ng,
                                         npw, -(-ng // mpi.world.size))
                bd2 = bg2.new_descriptor(npw, ng,
                                         -(-npw // mpi.world.size), ng)

                fhxc_Glr = np.zeros((len(l_pw_range), ng), dtype=complex)
                if ns == 2:
                    Koff_Glr = np.zeros((len(l_pw_range), ng), dtype=complex)

                r = Redistributor(bg1.comm, bd1, bd2)
                r.redistribute(fhxc_qsGr[iq][0], fhxc_Glr, npw, ng)
                if ns == 2:
                    r.redistribute(fhxc_qsGr[iq][1], Koff_Glr, npw, ng)
            else:
                fhxc_Glr = fhxc_qsGr[iq][0]
                if ns == 2:
                    Koff_Glr = fhxc_qsGr[iq][1]

            # Fourier transform of r'
            for iG in range(len(l_pw_range)):
                f_g = fhxc_Glr[iG].reshape(ng_c)
                f_G = self.pd.fft(f_g.conj(), iq) * vol / ng
                fhxc_sGsG[l_pw_range[0] + iG, :npw] = f_G.conj()
                if ns == 2:
                    v_g = Koff_Glr[iG].reshape(ng_c)
                    v_G = self.pd.fft(v_g.conj(), iq) * vol / ng
                    fhxc_sGsG[npw + l_pw_range[0] + iG, :npw] = v_G.conj()

            if ns == 2:  # f_00 = f_11 and f_01 = f_10
                fhxc_sGsG[:npw, npw:] = fhxc_sGsG[npw:, :npw]
                fhxc_sGsG[npw:, npw:] = fhxc_sGsG[:npw, :npw]

            mpi.world.sum(fhxc_sGsG)
            fhxc_sGsG /= vol

            if mpi.rank == 0:
                w = ulm.open('fhxc_%s_%s_%s_%s.ulm' %
                             (self.tag, self.xc, self.ecut, iq), 'w')
                if nR > 1:  # add Hartree kernel evaluated in PW basis
                    Gq2_G = self.pd.G2_qG[iq]
                    if (q == 0).all():
                        Gq2_G = Gq2_G.copy()
                        Gq2_G[0] = 1.
                    vq_G = 4 * np.pi / Gq2_G
                    fhxc_sGsG += np.tile(np.eye(npw) * vq_G, (ns, ns))
                w.write(fhxc_sGsG=fhxc_sGsG)
                w.close()
            mpi.world.barrier()
        print(file=self.fd)

    def calculate_local_kernel(self):
        # Standard ALDA exchange kernel
        # Use with care. Results are very difficult to converge
        # Sensitive to density_cut
        ns = self.calc.wfs.nspins
        gd = self.gd
        pd = self.pd
        cell_cv = gd.cell_cv
        icell_cv = 2 * np.pi * np.linalg.inv(cell_cv)
        vol = np.linalg.det(cell_cv)

        fxc_sg = ns * self.get_fxc_g(ns * self.n_g)
        fxc_sg[np.where(self.n_g < self.density_cut)] = 0.0

        r_vg = gd.get_grid_point_coordinates()

        for iq in range(len(self.ibzq_qc)):
            Gvec_Gc = np.dot(pd.get_reciprocal_vectors(q=iq, add_q=False),
                             cell_cv / (2 * np.pi))
            npw = len(Gvec_Gc)
            l_pw_size = -(-npw // mpi.world.size)
            l_pw_range = range(mpi.world.rank * l_pw_size,
                               min((mpi.world.rank + 1) * l_pw_size, npw))
            fhxc_sGsG = np.zeros((ns * npw, ns * npw), dtype=complex)
            for s in range(ns):
                for iG in l_pw_range:
                    for jG in range(npw):
                        fxc = fxc_sg[s].copy()
                        dG_c = Gvec_Gc[iG] - Gvec_Gc[jG]
                        dG_v = np.dot(dG_c, icell_cv)
                        dGr_g = gemmdot(dG_v, r_vg, beta=0.0)
                        ft_fxc = gd.integrate(np.exp(-1j * dGr_g) * fxc)
                        fhxc_sGsG[s * npw + iG, s * npw + jG] = ft_fxc

            mpi.world.sum(fhxc_sGsG)
            fhxc_sGsG /= vol

            Gq2_G = self.pd.G2_qG[iq]
            if (self.ibzq_qc[iq] == 0).all():
                Gq2_G[0] = 1.
            vq_G = 4 * np.pi / Gq2_G
            fhxc_sGsG += np.tile(np.eye(npw) * vq_G, (ns, ns))

            if mpi.rank == 0:
                w = ulm.open('fhxc_%s_%s_%s_%s.ulm' %
                             (self.tag, self.xc, self.ecut, iq), 'w')
                w.write(fhxc_sGsG=fhxc_sGsG)
                w.close()
            mpi.world.barrier()
        print(file=self.fd)

    def get_fxc_g(self, n_g, index=None):
        if self.xc[-3:] == 'LDA':
            return self.get_lda_g(n_g)
        elif self.xc[-3:] == 'PBE':
            return self.get_pbe_g(n_g, index=index)
        else:
            raise '%s kernel not recognized' % self.xc

    def get_lda_g(self, n_g):
        return (4. / 9.) * self.A_x * n_g**(-2. / 3.)

    def get_pbe_g(self, n_g, index=None):
        if index is None:
            gradn_vg = self.gradn_vg
        else:
            gradn_vg = self.calc.density.gd.empty(3)
            for v in range(3):
                gradn_vg[v] = (self.gradn_vg[v] +
                               self.gradn_vg[v].flatten()[index]) / 2

        kf_g = (3. * np.pi**2 * n_g)**(1 / 3.)
        s2_g = np.zeros_like(n_g)
        for v in range(3):
            axpy(1.0, gradn_vg[v]**2, s2_g)
        s2_g /= 4 * kf_g**2 * n_g**2

        e_g = self.A_x * n_g**(4 / 3.)
        v_g = (4 / 3.) * e_g / n_g
        f_g = (1 / 3.) * v_g / n_g

        kappa = 0.804
        mu = 0.2195149727645171

        denom_g = (1 + mu * s2_g / kappa)
        F_g = 1. + kappa - kappa / denom_g
        Fn_g = -mu / denom_g**2 * 8 * s2_g / (3 * n_g)
        Fnn_g = -11 * Fn_g / (3 * n_g) - 2 * Fn_g**2 / kappa

        fxc_g = f_g * F_g
        fxc_g += 2 * v_g * Fn_g
        fxc_g += e_g * Fnn_g

        """
        # Contributions from varying the gradient
        #Fgrad_vg = np.zeros_like(gradn_vg)
        #Fngrad_vg = np.zeros_like(gradn_vg)
        #for v in range(3):
        #    axpy(1.0, mu / den_g**2 * gradn_vg[v] / (2 * kf_g**2 * n_g**2),
        #         Fgrad_vg[v])
        #    axpy(-8.0, Fgrad_vg[v] / (3 * n_g), Fngrad_vg[v])
        #    axpy(-2.0, Fgrad_vg[v] * Fn_g / kappa, Fngrad_vg[v])

        #tmp = np.zeros_like(fxc_g)
        #tmp1 = np.zeros_like(fxc_g)

        #for v in range(3):
            #self.grad_v[v](Fgrad_vg[v], tmp)
            #axpy(-2.0, tmp * v_g, fxc_g)
            #for u in range(3):
                #self.grad_v[u](Fgrad_vg[u] * tmp, tmp1)
                #axpy(-4.0/kappa, tmp1 * e_g, fxc_g)
            #self.grad_v[v](Fngrad_vg[v], tmp)
            #axpy(-2.0, tmp * e_g, fxc_g)
        #self.laplace(mu / den_g**2 / (2 * kf_g**2 * n_g**2), tmp)
        #axpy(1.0, tmp * e_g, fxc_g)
        """

        return fxc_g


"""
    def get_fxc_libxc_g(self, n_g):
        ### NOT USED AT THE MOMENT
        gd = self.calc.density.gd.refine()

        xc = XC('GGA_X_' + self.xc[2:])
        #xc = XC('LDA_X')
        #sigma = np.zeros_like(n_g).flat[:]
        xc.set_grid_descriptor(gd)
        sigma_xg, gradn_svg = xc.calculate_sigma(np.array([n_g]))

        dedsigma_xg = np.zeros_like(sigma_xg)
        e_g = np.zeros_like(n_g)
        v_sg = np.array([np.zeros_like(n_g)])

        xc.calculate_gga(e_g, np.array([n_g]), v_sg, sigma_xg, dedsigma_xg)

        sigma = sigma_xg[0].flat[:]
        gradn_vg = gradn_svg[0]
        dedsigma_g = dedsigma_xg[0]

        libxc = LibXC('GGA_X_' + self.xc[2:])
        #libxc = LibXC('LDA_X')
        libxc.initialize(1)
        libxc_fxc = libxc.xc.calculate_fxc_spinpaired

        fxc_g = np.zeros_like(n_g).flat[:]
        d2edndsigma_g = np.zeros_like(n_g).flat[:]
        d2ed2sigma_g = np.zeros_like(n_g).flat[:]

        libxc_fxc(n_g.flat[:], fxc_g, sigma, d2edndsigma_g, d2ed2sigma_g)
        fxc_g = fxc_g.reshape(np.shape(n_g))
        d2edndsigma_g = d2edndsigma_g.reshape(np.shape(n_g))
        d2ed2sigma_g = d2ed2sigma_g.reshape(np.shape(n_g))

        tmp = np.zeros_like(fxc_g)
        tmp1 = np.zeros_like(fxc_g)

        #for v in range(3):
            #self.grad_v[v](d2edndsigma_g * gradn_vg[v], tmp)
            #axpy(-4.0, tmp, fxc_g)

        #for u in range(3):
            #for v in range(3):
                #self.grad_v[v](d2ed2sigma_g * gradn_vg[u] * gradn_vg[v], tmp)
                #self.grad_v[u](tmp, tmp1)
                #axpy(4.0, tmp1, fxc_g)

        #self.laplace(dedsigma_g, tmp)
        #axpy(2.0, tmp, fxc_g)

        return fxc_g[::2, ::2, ::2]

    def get_numerical_fxc_sg(self, n_sg):
        ### NOT USED AT THE MOMENT
        gd = self.calc.density.gd.refine()
        delta = 1.e-4

        if self.xc[2:] == 'LDA':
            xc = XC('LDA_X')
            v1xc_sg = np.zeros_like(n_sg)
            v2xc_sg = np.zeros_like(n_sg)
            xc.calculate(gd, (1 + delta) * n_sg, v1xc_sg)
            xc.calculate(gd, (1 - delta) * n_sg, v2xc_sg)
            fxc_sg = (v1xc_sg - v2xc_sg) / (2 * delta * n_sg)
        else:
            fxc_sg = np.zeros_like(n_sg)
            xc = XC('GGA_X_' + self.xc[2:])
            vxc_sg = np.zeros_like(n_sg)
            xc.calculate(gd, n_sg, vxc_sg)
            for s in range(len(n_sg)):
                for x in range(len(n_sg[0])):
                    for y in range(len(n_sg[0, 0])):
                        for z in range(len(n_sg[0, 0, 0])):
                            v1xc_sg = np.zeros_like(n_sg)
                            n1_sg = n_sg.copy()
                            n1_sg[s, x, y, z] *= (1 + delta)
                            xc.calculate(gd, n1_sg, v1xc_sg)
                            num = v1xc_sg[s, x, y, z] - vxc_sg[s, x, y, z]
                            den = delta * n_sg[s, x, y, z]
                            fxc_sg[s, x, y, z] = num / den

        return fxc_sg[:, ::2, ::2, ::2]
"""
def set_flags(self):
    """ Based on chosen fxc and av. scheme set up true-false flags """

    if self.xc not in ('RPA',
                       'range_RPA',  # range separated RPA a la Bruneval
                       'rALDA',      # renormalized kernels
                       'rAPBE',
                       'range_rALDA',
                       'rALDAns',    # no spin (ns)
                       'rAPBEns',
                       'rALDAc',     # rALDA + correlation
                       'CP',         # Constantin Pitarke
                       'CP_dyn',     # Dynamical form of CP
                       'CDOP',       # Corradini et al
                       'CDOPs',      # CDOP without local term
                       'JGMs',       # simplified jellium-with-gap kernel
                       'JGMsx',      # simplified jellium-with-gap kernel,
                                     # constructed with exchange part only
                                     # so that it scales linearly with l
                       'ALDA'        # standard ALDA
                       ):
        raise RuntimeError('%s kernel not recognized' % self.xc)

    if (self.xc == 'rALDA' or self.xc == 'rAPBE' or self.xc == 'ALDA'):

        if self.av_scheme is None:
            self.av_scheme = 'density'
            # Two-point scheme default for rALDA and rAPBE

        self.spin_kernel = True
        # rALDA/rAPBE are the only kernels which have spin-dependent forms

    else:
        self.spin_kernel = False

    if self.av_scheme == 'density':
        assert (self.xc == 'rALDA' or self.xc == 'rAPBE' or self.xc == 'ALDA'
                ), ('Two-point density average ' +
                    'only implemented for rALDA and rAPBE')

    elif self.xc not in ('RPA', 'range_RPA'):
        self.av_scheme = 'wavevector'
    else:
        self.av_scheme = None

    if self.xc in ('rALDAns', 'rAPBEns', 'range_RPA', 'JGMsx',
                   'RPA', 'rALDA', 'rAPBE', 'range_rALDA','ALDA'):
        self.linear_kernel = True  # Scales linearly with coupling constant
    else:
        self.linear_kernel = False

    if self.xc == 'CP_dyn':
        self.dyn_kernel = True
    else:
        self.dyn_kernel = False

    if self.xc == 'JGMs' or self.xc == 'JGMsx':
        assert (self.Eg is not None), 'JGMs kernel requires a band gap!'
        self.Eg /= Ha # Convert from eV
    else:
        self.Eg = None



from __future__ import print_function
import numpy as np
from gpaw.lfc import LocalizedFunctionsCollection as LFC
from gpaw.grid_descriptor import GridDescriptor
from gpaw.atom.radialgd import EquidistantRadialGridDescriptor
from gpaw.spline import Spline
from gpaw.setup import Setup

rc = 2.0
a = 2.5 * rc
n = 64
lmax = 2
b = 8.0
m = (lmax + 1)**2
gd = GridDescriptor([n, n, n], [a, a, a])
r = np.linspace(0, rc, 200)
g = np.exp(-(r / rc * b)**2)
splines = [Spline(l=l, rmax=rc, f_g=g) for l in range(lmax + 1)]
c = LFC(gd, [splines])
c.set_positions([(0, 0, 0)])
psi = gd.zeros(m)
d0 = c.dict(m)
if 0 in d0:
    d0[0] = np.identity(m)
c.add(psi, d0)
d1 = c.dict(m, derivative=True)
c.derivative(psi, d1)
class TestSetup(Setup):
    l_j = range(lmax + 1)
    nj = lmax + 1
    ni = m

    def __init__(self):
        pass
rgd = EquidistantRadialGridDescriptor(r[1], len(r))
g = [np.exp(-(r / rc * b)**2) * r**l for l in range(lmax + 1)]
d2 = TestSetup().get_derivative_integrals(rgd, g, np.zeros_like(g))
if 0 in d1:
    print(abs(d1[0] - d2).max())
    assert abs(d1[0] - d2).max() < 2e-6






"""Build GPAW's web-page.

Initial setup::

    cd ~
    python3 -m venv gpaw-web-page
    cd gpaw-web-page
    . bin/activate
    pip install sphinx-rtd-theme
    pip install Sphinx
    pip install matplotlib scipy
    git clone http://gitlab.com/ase/ase.git
    cd ase
    pip install -U .
    cd ..
    git clone http://gitlab.com/gpaw/gpaw.git
    cd gpaw
    python setup.py install

Crontab::

    WEB_PAGE_FOLDER=...
    CMD="python -m gpaw.utilities.build_web_page"
    10 20 * * * cd ~/gpaw-web-page; . bin/activate; cd gpaw; $CMD > ../gpaw.log

"""
from __future__ import print_function
import os
import subprocess
import sys

from gpaw import __version__


cmds = """\
touch ../gpaw-web-page.lock
cd ../ase; git checkout web-page -q; git pull -q; pip install .
git clean -fdx
git checkout web-page -q > /dev/null 2>&1
git pull -q > /dev/null 2>&1
python setup.py install {opts}
cd doc; sphinx-build -b html -d build/doctrees . build/html
mv doc/build/html gpaw-web-page
cd gpaw-web-page/_sources/setups; cp setups.rst.txt setups.txt
cd ../ase; git checkout master -q; pip install .
git clean -fdx doc
rm -r build
git checkout master -q > /dev/null 2>&1
git pull -q > /dev/null 2>&1
python setup.py install {opts}
cd doc; sphinx-build -b html -d build/doctrees . build/html
mv doc/build/html gpaw-web-page/dev
python setup.py sdist
cp dist/gpaw-*.tar.gz gpaw-web-page/
cp dist/gpaw-*.tar.gz gpaw-web-page/dev/
find gpaw-web-page -name install.html | xargs sed -i s/snapshot.tar.gz/{tar}/g
tar -czf gpaw-web-page.tar.gz gpaw-web-page
cp gpaw-web-page.tar.gz {dir}/tmp-gpaw-web-page.tar.gz
mv {dir}/tmp-gpaw-web-page.tar.gz {dir}/gpaw-web-page.tar.gz"""

cmds = cmds.format(opts=os.environ.get('GPAW_COMPILE_OPTIONS', ''),
                   tar='gpaw-' + __version__ + '.tar.gz',
                   dir=os.environ['WEB_PAGE_FOLDER'])


def build():
    if os.path.isfile('../gpaw-web-page.lock'):
        print('Locked', file=sys.stderr)
        return
    try:
        for cmd in cmds.splitlines():
            subprocess.check_call(cmd, shell=True)
    finally:
        os.remove('../gpaw-web-page.lock')


if __name__ == '__main__':
    build()



from __future__ import print_function
from ase import Atom, Atoms
from ase.parallel import parprint

from gpaw.test import equal
from gpaw import GPAW, mpi
from gpaw.lrtddft import LrTDDFT

txt = '-'
txt = '/dev/null'

load = False
    
if not load:
    R = 0.7  # approx. experimental bond length
    a = 3.0
    c = 4.0
    H2 = Atoms([Atom('H', (a / 2, a / 2, (c - R) / 2)),
                Atom('H', (a / 2, a / 2, (c + R) / 2))],
               cell=(a, a, c))
    calc = GPAW(xc='PBE', nbands=2, spinpol=False, txt=txt)
    H2.set_calculator(calc)
    H2.get_potential_energy()
else:
    calc = GPAW('H2.gpw', txt=txt)

# DFT only

xc = 'LDA'

# no spin

lr = LrTDDFT(calc, xc=xc)
lr.diagonalize()

lr_ApmB = LrTDDFT(calc, xc=xc, force_ApmB=True)
lr_ApmB.diagonalize()
parprint('lr=', lr)
parprint('ApmB=', lr_ApmB)
equal(lr[0].get_energy(), lr_ApmB[0].get_energy(), 5.e-9)

# with spin
parprint('------ with spin')

if not load:
    c_spin = GPAW(xc='PBE', nbands=2,
                  spinpol=True, parallel={'domain': mpi.world.size},
                  txt=txt)
    H2.set_calculator(c_spin)
    c_spin.calculate(H2)
else:
    c_spin = GPAW('H2spin.gpw', txt=txt)
lr = LrTDDFT(c_spin, xc=xc)
lr.diagonalize()

lr_ApmB = LrTDDFT(c_spin, xc=xc, force_ApmB=True)
lr_ApmB.diagonalize()
parprint('lr=', lr)
parprint('ApmB=', lr_ApmB)
equal(lr[0].get_energy(), lr_ApmB[0].get_energy(), 5.e-8)
equal(lr[1].get_energy(), lr_ApmB[1].get_energy(), 5.e-8)

# with spin virtual
parprint('------ with virtual spin')

lr = LrTDDFT(calc, xc=xc, nspins=2)
lr.diagonalize()

# ApmB
lr_ApmB = LrTDDFT(calc, xc=xc, nspins=2)
lr_ApmB.diagonalize()
parprint('lr=', lr)
parprint('ApmB=', lr_ApmB)
equal(lr[0].get_energy(), lr_ApmB[0].get_energy(), 5.e-8)
equal(lr[1].get_energy(), lr_ApmB[1].get_energy(), 5.e-8)
    
# with HF exchange

xc = 'PBE0'

parprint('------ with spin xc=', xc)
lr_spin = LrTDDFT(c_spin, xc=xc)
lr_spin.diagonalize()
parprint('lr=', lr_spin)

parprint('------ with virtual spin xc=', xc)
lr = LrTDDFT(calc, xc=xc, nspins=2)
lr.diagonalize()
parprint('lr=', lr)
equal(lr[0].get_energy(), lr_spin[0].get_energy(), 3.8e-6)
equal(lr[1].get_energy(), lr_spin[1].get_energy(), 3.4e-6)



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""
See Kresse, Phys. Rev. B 54, 11169 (1996)
"""

import numpy as np
from numpy.fft import fftn, ifftn

import gpaw.mpi as mpi
from gpaw.utilities.blas import axpy
from gpaw.fd_operators import FDOperator
from gpaw.utilities.tools import construct_reciprocal

"""About mixing-related classes.

(FFT/Broyden)BaseMixer: These classes know how to mix one density
array and store history etc.  But they do not take care of complexity
like spin.

(SpinSum/etc.)MixerDriver: These combine one or more BaseMixers to
implement a full algorithm.  Think of them as stateless (immutable).
The user can give an object of these types as input, but they will generally
be constructed by a utility function so the interface is nice.

The density object always wraps the (X)MixerDriver with a
MixerWrapper.  The wrapper contains the common code for all mixers so
we don't have to implement it multiple times (estimate memory, etc.).

In the end, what the user provides is probably a dictionary anyway, and the
relevant objects are instantiated automatically."""


class BaseMixer:
    name = 'pulay'

    """Pulay density mixer."""
    def __init__(self, beta, nmaxold, weight):
        """Construct density-mixer object.

        Parameters:

        beta: float
            Mixing parameter between zero and one (one is most
            aggressive).
        nmaxold: int
            Maximum number of old densities.
        weight: float
            Weight parameter for special metric (for long wave-length
            changes).

        """

        self.beta = beta
        self.nmaxold = nmaxold
        self.weight = weight

    def initialize_metric(self, gd):
        self.gd = gd

        if self.weight == 1:
            self.metric = None

        else:
            a = 0.125 * (self.weight + 7)
            b = 0.0625 * (self.weight - 1)
            c = 0.03125 * (self.weight - 1)
            d = 0.015625 * (self.weight - 1)
            self.metric = FDOperator([a,
                                      b, b, b, b, b, b,
                                      c, c, c, c, c, c, c, c, c, c, c, c,
                                      d, d, d, d, d, d, d, d],
                                     [(0, 0, 0),  # a
                                      (-1, 0, 0), (1, 0, 0),  # b
                                      (0, -1, 0), (0, 1, 0),
                                      (0, 0, -1), (0, 0, 1),
                                      (1, 1, 0), (1, 0, 1), (0, 1, 1),  # c
                                      (1, -1, 0), (1, 0, -1), (0, 1, -1),
                                      (-1, 1, 0), (-1, 0, 1), (0, -1, 1),
                                      (-1, -1, 0), (-1, 0, -1), (0, -1, -1),
                                      (1, 1, 1), (1, 1, -1), (1, -1, 1),  # d
                                      (-1, 1, 1), (1, -1, -1), (-1, -1, 1),
                                      (-1, 1, -1), (-1, -1, -1)],
                                     gd, float).apply
            self.mR_G = gd.empty()

    def reset(self):
        """Reset Density-history.

        Called at initialization and after each move of the atoms.

        my_nuclei:   All nuclei in local domain.
        """

        # History for Pulay mixing of densities:
        self.nt_iG = []  # Pseudo-electron densities
        self.R_iG = []  # Residuals
        self.A_ii = np.zeros((0, 0))

        self.D_iap = []
        self.dD_iap = []

    def calculate_charge_sloshing(self, R_G):
        return self.gd.integrate(np.fabs(R_G))

    def mix_single_density(self, nt_G, D_ap):
        iold = len(self.nt_iG)

        dNt = np.inf
        if iold > 0:
            if iold > self.nmaxold:
                # Throw away too old stuff:
                del self.nt_iG[0]
                del self.R_iG[0]
                del self.D_iap[0]
                del self.dD_iap[0]
                # for D_p, D_ip, dD_ip in self.D_a:
                #     del D_ip[0]
                #     del dD_ip[0]
                iold = self.nmaxold

            # Calculate new residual (difference between input and
            # output density):
            R_G = nt_G - self.nt_iG[-1]
            dNt = self.calculate_charge_sloshing(R_G)
            self.R_iG.append(R_G)
            self.dD_iap.append([])
            for D_p, D_ip in zip(D_ap, self.D_iap[-1]):
                self.dD_iap[-1].append(D_p - D_ip)

            # Update matrix:
            A_ii = np.zeros((iold, iold))
            i2 = iold - 1

            if self.metric is None:
                mR_G = R_G
            else:
                mR_G = self.mR_G
                self.metric(R_G, mR_G)

            for i1, R_1G in enumerate(self.R_iG):
                a = self.gd.comm.sum(self.dotprod(R_1G, mR_G, self.dD_iap[i1],
                                                  self.dD_iap[-1]))
                A_ii[i1, i2] = a
                A_ii[i2, i1] = a
            A_ii[:i2, :i2] = self.A_ii[-i2:, -i2:]
            self.A_ii = A_ii

            try:
                B_ii = np.linalg.inv(A_ii)
            except np.linalg.LinAlgError:
                alpha_i = np.zeros(iold)
                alpha_i[-1] = 1.0
            else:
                alpha_i = B_ii.sum(1)
                try:
                    # Normalize:
                    alpha_i /= alpha_i.sum()
                except ZeroDivisionError:
                    alpha_i[:] = 0.0
                    alpha_i[-1] = 1.0

            # Calculate new input density:
            nt_G[:] = 0.0
            # for D_p, D_ip, dD_ip in self.D_a:
            for D in D_ap:
                D[:] = 0.0
            beta = self.beta
            for i, alpha in enumerate(alpha_i):
                axpy(alpha, self.nt_iG[i], nt_G)
                axpy(alpha * beta, self.R_iG[i], nt_G)
                for D_p, D_ip, dD_ip in zip(D_ap, self.D_iap[i],
                                            self.dD_iap[i]):
                    axpy(alpha, D_ip, D_p)
                    axpy(alpha * beta, dD_ip, D_p)

        # Store new input density (and new atomic density matrices):
        self.nt_iG.append(nt_G.copy())
        self.D_iap.append([])
        for D_p in D_ap:
            self.D_iap[-1].append(D_p.copy())
        return dNt

    # may presently be overridden by passing argument in constructor
    def dotprod(self, R1_G, R2_G, dD1_ap, dD2_ap):
        return np.vdot(R1_G, R2_G).real

    def estimate_memory(self, mem, gd):
        gridbytes = gd.bytecount()
        mem.subnode('nt_iG, R_iG', 2 * self.nmaxold * gridbytes)

    def __repr__(self):
        classname = self.__class__.__name__
        template = '%s(beta=%f, nmaxold=%d, weight=%f)'
        string = template % (classname, self.beta, self.nmaxold, self.weight)
        return string


class ExperimentalDotProd:
    def __init__(self, calc):
        self.calc = calc

    def __call__(self, R1_G, R2_G, dD1_ap, dD2_ap):
        prod = np.vdot(R1_G, R2_G).real
        setups = self.calc.wfs.setups
        # okay, this is a bit nasty because it depends on dD1_ap
        # and its friend having come from D_asp.values() and the dictionaries
        # not having been modified.  This is probably true... for now.
        avalues = self.calc.density.D_asp.keys()
        for a, dD1_p, dD2_p in zip(avalues, dD1_ap, dD2_ap):
            I4_pp = setups[a].four_phi_integrals()
            dD4_pp = np.outer(dD1_p, dD2_p)  # not sure if corresponds quite
            prod += (I4_pp * dD4_pp).sum()
        return prod


class ReciprocalMetric:
    def __init__(self, weight, k2_Q):
        self.k2_Q = k2_Q
        k2_min = np.min(self.k2_Q)
        self.q1 = (weight - 1) * k2_min

    def __call__(self, R_Q, mR_Q):
        mR_Q[:] = R_Q * (1.0 + self.q1 / self.k2_Q)


class FFTBaseMixer(BaseMixer):
    name = 'fft'

    """Mix the density in Fourier space"""
    def __init__(self, beta, nmaxold, weight):
        BaseMixer.__init__(self, beta, nmaxold, weight)
        self.gd1 = None

    def initialize_metric(self, gd):
        self.gd = gd

        if gd.comm.rank == 0:
            self.gd1 = gd.new_descriptor(comm=mpi.serial_comm)
            k2_Q, _ = construct_reciprocal(self.gd1)
            self.metric = ReciprocalMetric(self.weight, k2_Q)
            self.mR_G = self.gd1.empty(dtype=complex)
        else:
            self.metric = lambda R_Q, mR_Q: None
            self.mR_G = np.empty((0, 0, 0), dtype=complex)

    def calculate_charge_sloshing(self, R_Q):
        if self.gd.comm.rank == 0:
            cs = self.gd1.integrate(np.fabs(ifftn(R_Q).real))
        else:
            cs = 0.0
        return self.gd.comm.sum(cs)

    def mix_single_density(self, nt_G, D_ap):
        # Transform real-space density to Fourier space
        nt1_G = self.gd.collect(nt_G)
        if self.gd.comm.rank == 0:
            nt_Q = np.ascontiguousarray(fftn(nt1_G))
        else:
            nt_Q = np.empty((0, 0, 0), dtype=complex)

        dNt = BaseMixer.mix_single_density(self, nt_Q, D_ap)

        # Return density in real space
        if self.gd.comm.rank == 0:
            nt1_G = ifftn(nt_Q).real
        self.gd.distribute(nt1_G, nt_G)

        return dNt


class BroydenBaseMixer:
    name = 'broyden'

    def __init__(self, beta, nmaxold, weight):
        self.verbose = False
        self.beta = beta
        self.nmaxold = nmaxold
        self.weight = 1.0  # XXX discards argument

    def initialize_metric(self, gd):
        self.gd = gd

    def reset(self):
        self.step = 0
        # self.d_nt_G = []
        # self.d_D_ap = []

        self.R_iG = []
        self.dD_iap = []

        self.nt_iG = []
        self.D_iap = []
        self.c_G = []
        self.v_G = []
        self.u_G = []
        self.u_D = []

    def mix_single_density(self, nt_G, D_ap):
        dNt = np.inf
        if self.step > 2:
            del self.R_iG[0]
            for d_Dp in self.dD_iap:
                del d_Dp[0]
        if self.step > 0:
            self.R_iG.append(nt_G - self.nt_iG[-1])
            for d_Dp, D_p, D_ip in zip(self.dD_iap, D_ap, self.D_iap):
                d_Dp.append(D_p - D_ip[-1])
            fmin_G = self.gd.integrate(self.R_iG[-1] * self.R_iG[-1])
            dNt = self.gd.integrate(np.fabs(self.R_iG[-1]))
            if self.verbose:
                print('Mixer: broyden: fmin_G = %f fmin_D = %f' % fmin_G)
        if self.step == 0:
            self.eta_G = np.empty(nt_G.shape)
            self.eta_D = []
            for D_p in D_ap:
                self.eta_D.append(0)
                self.u_D.append([])
                self.D_iap.append([])
                self.dD_iap.append([])
        else:
            if self.step >= 2:
                del self.c_G[:]
                if len(self.v_G) >= self.nmaxold:
                    del self.u_G[0]
                    del self.v_G[0]
                    for u_D in self.u_D:
                        del u_D[0]
                temp_nt_G = self.R_iG[1] - self.R_iG[0]
                self.v_G.append(temp_nt_G / self.gd.integrate(temp_nt_G *
                                                              temp_nt_G))
                if len(self.v_G) < self.nmaxold:
                    nstep = self.step - 1
                else:
                    nstep = self.nmaxold
                for i in range(nstep):
                    self.c_G.append(self.gd.integrate(self.v_G[i] *
                                                      self.R_iG[1]))
                self.u_G.append(self.beta * temp_nt_G + self.nt_iG[1] -
                                self.nt_iG[0])
                for d_Dp, u_D, D_ip in zip(self.dD_iap, self.u_D, self.D_iap):
                    temp_D_ap = d_Dp[1] - d_Dp[0]
                    u_D.append(self.beta * temp_D_ap + D_ip[1] - D_ip[0])
                usize = len(self.u_G)
                for i in range(usize - 1):
                    a_G = self.gd.integrate(self.v_G[i] * temp_nt_G)
                    axpy(-a_G, self.u_G[i], self.u_G[usize - 1])
                    for u_D in self.u_D:
                        axpy(-a_G, u_D[i], u_D[usize - 1])
            self.eta_G = self.beta * self.R_iG[-1]
            for i, d_Dp in enumerate(self.dD_iap):
                self.eta_D[i] = self.beta * d_Dp[-1]
            usize = len(self.u_G)
            for i in range(usize):
                axpy(-self.c_G[i], self.u_G[i], self.eta_G)
                for eta_D, u_D in zip(self.eta_D, self.u_D):
                    axpy(-self.c_G[i], u_D[i], eta_D)
            axpy(-1.0, self.R_iG[-1], nt_G)
            axpy(1.0, self.eta_G, nt_G)
            for D_p, d_Dp, eta_D in zip(D_ap, self.dD_iap, self.eta_D):
                axpy(-1.0, d_Dp[-1], D_p)
                axpy(1.0, eta_D, D_p)
            if self.step >= 2:
                del self.nt_iG[0]
                for D_ip in self.D_iap:
                    del D_ip[0]
        self.nt_iG.append(np.copy(nt_G))
        for D_ip, D_p in zip(self.D_iap, D_ap):
            D_ip.append(np.copy(D_p))
        self.step += 1
        return dNt


class DummyMixer:
    """Dummy mixer for TDDFT, i.e., it does not mix."""
    name = 'dummy'
    beta = 1.0
    nmaxold = 1
    weight = 1

    def mix(self, basemixers, nt_sG, D_asp):
        return 0.0

    def get_basemixers(self, nspins):
        return []

    def todict(self):
        return {'name': 'dummy'}


class SeparateSpinMixerDriver:
    name = 'separate'

    def __init__(self, basemixerclass, beta, nmaxold, weight):
        self.basemixerclass = basemixerclass

        self.beta = beta
        self.nmaxold = nmaxold
        self.weight = weight

    def get_basemixers(self, nspins):
        return [self.basemixerclass(self.beta, self.nmaxold, self.weight)
                for _ in range(nspins)]

    def mix(self, basemixers, nt_sG, D_asp):
        """Mix pseudo electron densities."""
        D_asp = D_asp.values()
        D_sap = []
        for s in range(len(nt_sG)):
            D_sap.append([D_sp[s] for D_sp in D_asp])
        dNt = 0.0
        for nt_G, D_ap, basemixer in zip(nt_sG, D_sap, basemixers):
            dNt += basemixer.mix_single_density(nt_G, D_ap)
        return dNt


class SpinSumMixerDriver:
    name = 'sum'
    mix_atomic_density_matrices = False

    def __init__(self, basemixerclass, beta, nmaxold, weight):
        self.basemixerclass = basemixerclass

        self.beta = beta
        self.nmaxold = nmaxold
        self.weight = weight

    def get_basemixers(self, nspins):
        if nspins == 1:
            raise ValueError('Spin sum mixer expects 2 or 4 components')
        return [self.basemixerclass(self.beta, self.nmaxold, self.weight)]

    def mix(self, basemixers, nt_sG, D_asp):
        assert len(basemixers) == 1
        basemixer = basemixers[0]
        D_asp = D_asp.values()

        collinear = len(nt_sG) == 2

        # Mix density
        if collinear:
            nt_G = nt_sG.sum(0)
        else:
            nt_G = nt_sG[0]

        if self.mix_atomic_density_matrices:
            if collinear:
                D_ap = [D_sp[0] + D_sp[1] for D_sp in D_asp]
            else:
                D_ap = [D_sp[0] for D_sp in D_asp]
            dNt = basemixer.mix_single_density(nt_G, D_ap)
            if collinear:
                dD_ap = [D_sp[0] - D_sp[1] for D_sp in D_asp]
                for D_sp, D_p, dD_p in zip(D_asp, D_ap, dD_ap):
                    D_sp[0] = 0.5 * (D_p + dD_p)
                    D_sp[1] = 0.5 * (D_p - dD_p)
        else:
            dNt = basemixer.mix_single_density(nt_G, D_asp)

        if collinear:
            dnt_G = nt_sG[0] - nt_sG[1]
            # Only new magnetization for spin density
            # dD_ap = [D_sp[0] - D_sp[1] for D_sp in D_asp]

            # Construct new spin up/down densities
            nt_sG[0] = 0.5 * (nt_G + dnt_G)
            nt_sG[1] = 0.5 * (nt_G - dnt_G)

        return dNt


class SpinSumMixerDriver2(SpinSumMixerDriver):
    name = 'sum2'
    mix_atomic_density_matrices = True


class SpinDifferenceMixerDriver:
    name = 'difference'

    def __init__(self, basemixerclass, beta, nmaxold, weight,
                 beta_m=0.7, nmaxold_m=2, weight_m=10.0):
        self.basemixerclass = basemixerclass
        self.beta = beta
        self.nmaxold = nmaxold
        self.weight = weight
        self.beta_m = beta_m
        self.nmaxold_m = nmaxold_m
        self.weight_m = weight_m

    def get_basemixers(self, nspins):
        if nspins == 1:
            raise ValueError('Spin difference mixer expects 2 or 4 components')
        basemixer = self.basemixerclass(self.beta, self.nmaxold, self.weight)
        if nspins == 2:
            basemixer_m = self.basemixerclass(self.beta_m, self.nmaxold_m,
                                              self.weight_m)
            return basemixer, basemixer_m
        else:
            basemixer_x = self.basemixerclass(self.beta_m, self.nmaxold_m,
                                              self.weight_m)
            basemixer_y = self.basemixerclass(self.beta_m, self.nmaxold_m,
                                              self.weight_m)
            basemixer_z = self.basemixerclass(self.beta_m, self.nmaxold_m,
                                              self.weight_m)
            return basemixer, basemixer_x, basemixer_y, basemixer_z

    def mix(self, basemixers, nt_sG, D_asp):
        D_asp = D_asp.values()

        if len(nt_sG) == 2:
            basemixer, basemixer_m = basemixers
        else:
            assert len(nt_sG) == 4
            basemixer, basemixer_x, basemixer_y, basemixer_z = basemixers

        if len(nt_sG) == 2:
            # Mix density
            nt_G = nt_sG.sum(0)
            D_ap = [D_sp[0] + D_sp[1] for D_sp in D_asp]
            dNt = basemixer.mix_single_density(nt_G, D_ap)

            # Mix magnetization
            dnt_G = nt_sG[0] - nt_sG[1]
            dD_ap = [D_sp[0] - D_sp[1] for D_sp in D_asp]
            basemixer_m.mix_single_density(dnt_G, dD_ap)
            # (The latter is not counted in dNt)

            # Construct new spin up/down densities
            nt_sG[0] = 0.5 * (nt_G + dnt_G)
            nt_sG[1] = 0.5 * (nt_G - dnt_G)
            for D_sp, D_p, dD_p in zip(D_asp, D_ap, dD_ap):
                D_sp[0] = 0.5 * (D_p + dD_p)
                D_sp[1] = 0.5 * (D_p - dD_p)
        else:
            # Mix density
            nt_G = nt_sG[0]
            D_ap = [D_sp[0] for D_sp in D_asp]
            dNt = basemixer.mix_single_density(nt_G, D_ap)

            # Mix magnetization
            Dx_ap = [D_sp[1] for D_sp in D_asp]
            Dy_ap = [D_sp[2] for D_sp in D_asp]
            Dz_ap = [D_sp[3] for D_sp in D_asp]
            basemixer_x.mix_single_density(nt_sG[1], Dx_ap)
            basemixer_y.mix_single_density(nt_sG[2], Dy_ap)
            basemixer_z.mix_single_density(nt_sG[3], Dz_ap)

        return dNt


# Dictionaries to get mixers by name:
_backends = {}
_methods = {}
for cls in [FFTBaseMixer, BroydenBaseMixer, BaseMixer]:
    _backends[cls.name] = cls
for cls in [SeparateSpinMixerDriver, SpinSumMixerDriver, SpinSumMixerDriver2,
            SpinDifferenceMixerDriver, DummyMixer]:
    _methods[cls.name] = cls


# This function is used by Density to decide mixer parameters
# that the user did not explicitly provide, i.e., it fills out
# everything that is missing and returns a mixer "driver".
def get_mixer_from_keywords(pbc, nspins, **mixerkwargs):
    # The plan is to first establish a kwargs dictionary with all the
    # defaults, then we update it with values from the user.
    kwargs = {'backend': BaseMixer}

    if np.any(pbc):  # Works on array or boolean
        kwargs.update(beta=0.05, history=5, weight=50.0)
    else:
        kwargs.update(beta=0.25, history=3, weight=1.0)

    if nspins == 1:
        kwargs['method'] = SeparateSpinMixerDriver
    else:
        kwargs['method'] = SpinSumMixerDriver

    # Clean up mixerkwargs (compatibility)
    if 'nmaxold' in mixerkwargs:
        assert 'history' not in mixerkwargs
        mixerkwargs['history'] = mixerkwargs.pop('nmaxold')

    # Now the user override:
    for key in kwargs:
        # Clean any 'None' values out as if they had never been passed:
        val = mixerkwargs.pop(key, None)
        if val is not None:
            kwargs[key] = val

    # Resolve keyword strings (like 'fft') into classes (like FFTBaseMixer):
    driver = _methods.get(kwargs['method'], kwargs['method'])
    baseclass = _backends.get(kwargs['backend'], kwargs['backend'])

    # We forward any remaining mixer kwargs to the actual mixer object.
    # Any user defined variables that do not really exist will cause an error.
    mixer = driver(baseclass, beta=kwargs['beta'],
                   nmaxold=kwargs['history'], weight=kwargs['weight'],
                   **mixerkwargs)
    return mixer


# This is the only object which will be used by Density, sod the others
class MixerWrapper:
    def __init__(self, driver, nspins, gd):
        self.driver = driver

        self.beta = driver.beta
        self.nmaxold = driver.nmaxold
        self.weight = driver.weight
        assert self.weight is not None, driver

        self.basemixers = self.driver.get_basemixers(nspins)
        for basemixer in self.basemixers:
            basemixer.initialize_metric(gd)

    def mix(self, nt_sG, D_asp):
        return self.driver.mix(self.basemixers, nt_sG, D_asp)

    def estimate_memory(self, mem, gd):
        for i, basemixer in enumerate(self.basemixers):
            basemixer.estimate_memory(mem.subnode('Mixer %d' % i), gd)

    def reset(self):
        for basemixer in self.basemixers:
            basemixer.reset()

    def __str__(self):
        lines = ['Density mixing:',
                 'Method: ' + self.driver.name,
                 'Backend: ' + self.driver.basemixerclass.name,
                 'Linear mixing parameter: %g' % self.beta,
                 'Mixing with %d old densities' % self.nmaxold]
        if self.weight == 1:
            lines.append('No damping of long wave oscillations')
        else:
            lines.append('Damping of long wave oscillations: %g' % self.weight)
        return '\n  '.join(lines)


# Helper function to define old-style interfaces to mixers.
# Defines and returns a function which looks like a mixer class
def _definemixerfunc(method, backend):
    def getmixer(beta=None, nmaxold=None, weight=None, **kwargs):
        d = dict(method=method, backend=backend,
                 beta=beta, nmaxold=nmaxold, weight=weight)
        d.update(kwargs)
        return d
    return getmixer


Mixer = _definemixerfunc('separate', 'pulay')
MixerSum = _definemixerfunc('sum', 'pulay')
MixerSum2 = _definemixerfunc('sum2', 'pulay')
MixerDif = _definemixerfunc('difference', 'pulay')
FFTMixer = _definemixerfunc('separate', 'fft')
FFTMixerSum = _definemixerfunc('sum', 'fft')
FFTMixerSum2 = _definemixerfunc('sum2', 'fft')
FFTMixerDif = _definemixerfunc('difference', 'fft')
BroydenMixer = _definemixerfunc('separate', 'broyden')
BroydenMixerSum = _definemixerfunc('sum', 'broyden')
BroydenMixerSum2 = _definemixerfunc('sum2', 'broyden')
BroydenMixerDif = _definemixerfunc('difference', 'broyden')



from __future__ import print_function
import numpy as np
from time import time

from gpaw.utilities.lapack import diagonalize, diagonalize_mr3

seed = 43
gen = np.random.RandomState(seed)

def main(i,seed=42,dtype=float):
    if (dtype==complex):
        epsilon = 0.1j
    else:
        epsilon = 0.1
    x = i + 1
    N = x*100
    print("N =",N)
    H0 = np.zeros((N,N),dtype=dtype) + gen.rand(*(N,N))
    H1 = H0 + epsilon*np.tri(N,N, k=-1)
    W0 = np.zeros((N))
    Z0 = np.zeros_like(H0)
    t0 = time()
    diagonalize(H1, W0)
    t1 = time() - t0
    print("diagonalize", t1)
    t2 = time() 
    diagonalize_mr3(H1, W0, Z0)
    t3 = time() - t2
    print("diagonalize_mr3",t3)
    # diagonalize_mr3 must be faster than diagonalize 
    assert(t3 < t1)

if __name__ in ['__main__', '__builtin__']:
    for i in range(8): # Check matrix sizes only up to 800
        main(i,dtype=float)
        main(i,dtype=complex)



import numpy as np
from ase.build import bulk
from gpaw import GPAW, PW, Mixer
from gpaw.mpi import world

xc = 'PBE'
si = bulk('Si')
k = 3
si.calc = GPAW(mode=PW(250),
               mixer=Mixer(0.7, 5, 50.0),
               xc=xc,
               kpts=(k, k, k),
               convergence={'energy': 1e-8},
               parallel={'domain': min(2, world.size)},
               txt='si.txt')

si.set_cell(np.dot(si.cell,
                   [[1.02, 0, 0.03],
                    [0, 0.99, -0.02],
                    [0.2, -0.01, 1.03]]),
            scale_atoms=True)

e = si.get_potential_energy()

# Trigger nasty bug (fixed in !486):
si.calc.wfs.pt.blocksize = si.calc.wfs.pd.maxmyng - 1

s_analytical = si.get_stress()
s_numerical = si.calc.calculate_numerical_stress(si, 1e-5)
s_err = s_numerical - s_analytical

print('Analytical stress:\n', s_analytical)
print('Numerical stress:\n', s_numerical)
print('Error in stress:\n', s_err)
assert np.all(abs(s_err) < 1e-4)



"""Test GW band-gaps for Si."""

from ase.build import bulk
import numpy as np

from gpaw.test import equal
from gpaw import GPAW, PW, FermiDirac
from gpaw.response.g0w0 import G0W0


def run(atoms, symm, name):
    atoms.calc = GPAW(mode=PW(250),
                      eigensolver='rmm-diis',
                      occupations=FermiDirac(0.01),
                      symmetry=symm,
                      kpts={'size': (2, 2, 2), 'gamma': True},
                      parallel={'domain': 1},
                      txt=name + '.txt')
    e = atoms.get_potential_energy()
    scalapack = atoms.calc.wfs.bd.comm.size
    atoms.calc.diagonalize_full_hamiltonian(nbands=8, scalapack=scalapack)
    atoms.calc.write(name + '.gpw', mode='all')
    gw = G0W0(name + '.gpw', 'gw-' + name,
              nbands=8,
              integrate_gamma=0,
              kpts=[(0, 0, 0), (0.5, 0.5, 0)],  # Gamma, X
              ecut=40,
              domega0=0.1,
              eta=0.2,
              relbands=(-1, 2))  # homo, lumo, lumo+1, same as bands=(3, 6)
    results = gw.calculate()
    return e, results


a = 5.43
si1 = bulk('Si', 'diamond', a=a)
si2 = si1.copy()
si2.positions -= a / 8

i = 0
results = []
for si in [si1, si2]:
    for symm in [{}, 'off', {'time_reversal': False}, {'point_group': False}]:
        e, r = run(si, symm, str(i))
        G, X = r['eps'][0]
        results.append([e, G[0], G[1] - G[0], X[1] - G[0], X[2] - X[1]])
        G, X = r['qp'][0]
        results[-1].extend([G[0], G[1] - G[0], X[1] - G[0], X[2] - X[1]])
        i += 1

equal(abs(np.array(results[0]) -
          [-9.25,
           5.44, 2.39, 0.40, 0,
           6.26, 3.57, 1.32, 0]).max(), 0, 0.025)
equal(np.ptp(results, 0).max(), 0, 0.005)



"""Test correctness of vdW-DF potential."""
from math import pi
from gpaw.grid_descriptor import GridDescriptor
import numpy as np
from gpaw.test import equal
from gpaw.xc import XC
from gpaw.mpi import world

N = 10
a = 2.0
gd = GridDescriptor((N, N, N), (a, a, a))


def paired():
    xc = XC('vdW-DF')
    n = 0.3 * np.ones((1, N, N, N))
    n += 0.01 * np.cos(np.arange(N) * 2 * pi / N)
    v = 0.0 * n
    xc.calculate(gd, n, v)
    n2 = 1.0 * n
    i = 1
    n2[0, i, i, i] += 0.00002
    x = v[0, i, i, i] * gd.dv
    E2 = xc.calculate(gd, n2, v)
    n2[0, i, i, i] -= 0.00004
    E2 -= xc.calculate(gd, n2, v)
    x2 = E2 / 0.00004
    print(i, x, x2, x - x2, x / x2)
    equal(x, x2, 2e-11)


def polarized():
    xc = XC('vdW-DF')
    n = 0.04 * np.ones((2, N, N, N))
    n[1] = 0.3
    n[0] += 0.02 * np.sin(np.arange(N) * 2 * pi / N)
    n[1] += 0.2 * np.cos(np.arange(N) * 2 * pi / N)
    v = 0.0 * n
    xc.calculate(gd, n, v)
    n2 = 1.0 * n
    i = 1
    n2[0, i, i, i] += 0.00002
    x = v[0, i, i, i] * gd.dv
    E2 = xc.calculate(gd, n2, v)
    n2[0, i, i, i] -= 0.00004
    E2 -= xc.calculate(gd, n2, v)
    x2 = E2 / 0.00004
    print(i, x, x2, x - x2, x / x2)
    equal(x, x2, 2e-10)


if world.size == 1:
    polarized()
    paired()



import numpy as np

from ase import Atoms

from gpaw import GPAW, FermiDirac
from gpaw.utilities import compiled_with_sl
from gpaw.wavefunctions.pw import PW
from gpaw.response.df import DielectricFunction
from gpaw.mpi import world

# This test assures that some things that
# should be equal, are.

a = 2.5
c = 3.22

GR = Atoms(symbols='C2',
           positions=[(0.5 * a, 0.2 - np.sqrt(3) / 6 * a, 0.0),
                      (0.5 * a, 0.2 + np.sqrt(3) / 6 * a, 0.0)],
           cell=[(0.5 * a, -0.5 * 3**0.5 * a, 0),
                 (0.5 * a, 0.5 * 3**0.5 * a, 0),
                 (0.0, 0.0, c * 2.0)])
GR.set_pbc((True, True, True))
atoms = GR
GSsettings = [{'symmetry': 'off', 'kpts': {'density': 2.5, 'gamma': False}},
              {'symmetry': {}, 'kpts': {'density': 2.5, 'gamma': False}},
              {'symmetry': 'off', 'kpts': {'density': 2.5, 'gamma': True}},
              {'symmetry': {}, 'kpts': {'density': 2.5, 'gamma': True}}]

DFsettings = [{'disable_point_group': True,
               'disable_time_reversal': True},
              {'disable_point_group': False,
               'disable_time_reversal': True},
              {'disable_point_group': True,
               'disable_time_reversal': False},
              {'disable_point_group': False,
               'disable_time_reversal': False}]

if world.size > 1 and compiled_with_sl():
    DFsettings.append({'disable_point_group': False,
                       'disable_time_reversal': False,
                       'nblocks': 2})
              
for GSkwargs in GSsettings:
    calc = GPAW(h=0.18,
                mode=PW(600),
                occupations=FermiDirac(0.2),
                **GSkwargs)
 
    atoms.set_calculator(calc)
    atoms.get_potential_energy()
    calc.write('gr.gpw', 'all')

    dfs = []
    for kwargs in DFsettings:
        DF = DielectricFunction(calc='gr.gpw',
                                domega0=0.2,
                                eta=0.2,
                                ecut=40.0,
                                **kwargs)
        df1, df2 = DF.get_dielectric_function()
        if world.rank == 0:
            dfs.append(df1)

    # Check the calculated dielectric functions against
    # each other.
    while len(dfs):
        df = dfs.pop()
        for DFkwargs, df2 in zip(DFsettings[-len(dfs):], dfs):
            try:
                assert np.allclose(df, df2)
                print('Ground state settings:', GSkwargs)
                print('DFkwargs1:', DFsettings[-len(dfs) - 1])
                print('DFkwargs2:', DFkwargs)
                print(np.max(np.abs((df - df2) / df)))
            except AssertionError:
                print('Some symmetry or block-par. related problems')
                print('for calculation with following ground state settings')
                print('Ground state settings:', GSkwargs)
                print('The following DF settings do not return the ' +
                      'same results')
                print('DFkwargs1:', DFsettings[-len(dfs) - 1])
                print('DFkwargs2:', DFkwargs)
                print(np.max(np.abs((df - df2) / df)))
                raise AssertionError




import numpy as np
from ase.build import bulk
from gpaw import GPAW, FermiDirac, Mixer

atoms = bulk('Si', 'diamond', a=5.4834322363595565)
atoms *= (3, 3, 3)
atoms.calc = GPAW(
    gpts=(32, 32, 32),
    mixer=Mixer(0.5, 5, 50.0),
    txt='grumble.txt',
    kpts=(2, 1, 1),
    mode='lcao',
    basis='sz(dzp)',
    xc='oldLDA',
    occupations=FermiDirac(0.01))
f = atoms.get_forces()
fmax = np.abs(f).max()
print('maxforce', fmax)
assert fmax < 0.05  # 0.03 normally, 2.7 with bug
# fmax can be converged much closer to 0 with better grid/SCF convergence.




from ase import Atoms
from gpaw import GPAW, PW
from gpaw.mpi import world

if world.size == 1:
    atoms = Atoms('H', cell=(2, 2, 2), pbc=True)
    atoms.calc = GPAW(mode=PW(300, force_complex_dtype=True),
                      eigensolver='direct')
    atoms.get_potential_energy()

if world.size == 2:
    atoms = Atoms('H', cell=(2, 2, 2), pbc=True)
    atoms.calc = GPAW(mode=PW(300),
                      eigensolver='direct',
                      kpts=(3, 2, 2))
    atoms.get_potential_energy()



from ase import Atoms
from ase.units import Hartree
from gpaw import GPAW

h2 = Atoms('H2', [(0, 0, 0), (0, 0, 1)])
h2.center(vacuum=2.0)
calc = GPAW(nbands=2, convergence={'eigenstates': 1e-3})
h2.set_calculator(calc)
e0 = h2.get_potential_energy()
r0 = calc.wfs.eigensolver.error * Hartree**2 / 2
assert r0 < 1e-3
calc.write('h2', 'all')
    
# refine the restart file containing the wfs
calc = GPAW('h2', convergence={'eigenstates': 1e-5})
e1 = calc.get_atoms().get_potential_energy()
r1 = calc.wfs.eigensolver.error * Hartree**2 / 2
assert r1 < 1e-5



from gpaw import GPAW
from gpaw.cluster import Cluster
from gpaw.test import equal
from ase.build import molecule
from ase.units import mol, kcal
from gpaw.solvation import (
    SolvationGPAW,
    get_HW14_water_kwargs,
)

SKIP_VAC_CALC = True

h = 0.24
vac = 4.0
convergence = {
    'energy': 0.05 / 8.,
    'density': 10.,
    'eigenstates': 10.,
}

atoms = Cluster(molecule('H2O'))
atoms.minimal_box(vac, h)

if not SKIP_VAC_CALC:
    atoms.calc = GPAW(xc='PBE', h=h, convergence=convergence)
    Evac = atoms.get_potential_energy()
    print(Evac)
else:
    # h=0.24, vac=4.0, setups: 0.9.11271, convergence: only energy 0.05 / 8
    Evac = -14.857414548

atoms.calc = SolvationGPAW(
    xc='PBE', h=h, convergence=convergence,
    **get_HW14_water_kwargs()
)
Ewater = atoms.get_potential_energy()
Eelwater = atoms.calc.get_electrostatic_energy()
Esurfwater = atoms.calc.get_solvation_interaction_energy('surf')
atoms.get_forces()
DGSol = (Ewater - Evac) / (kcal / mol)
print('Delta Gsol: %s kcal / mol' % DGSol)

equal(DGSol, -6.3, 2.)
equal(Ewater, Eelwater + Esurfwater, 1e-14)



from optparse import OptionParser
import numpy as np

usage = '%prog [OPTION] FILE'
description = 'Parse test results in FILE and reorder tests by duration.'

p = OptionParser(usage=usage,
                 description=description)
p.add_option('--pretty', action='store_true',
             help='print nicely with timings.  '
             'Otherwise print Python syntax')
opts, args = p.parse_args()
if len(args) != 1:
    p.error('One file please.')

fd = open(args[0])
for line in fd:
    if line.startswith('===='):
        break

tests = []
durations = []
results = []

for line in fd:
    if line.startswith('===='):
        break
    t1, t2, t3 = line.split()
    tests.append(t1)
    if t3 == 'OK':
        durations.append(float(t2))
    else:
        durations.append(np.inf)
    results.append(t3)

# Use stable sort for minimum shuffling
args = np.argsort(durations, kind='mergesort')
tests = np.array(tests)[args]
durations = np.array(durations)[args]
results = np.array(results)[args]

if opts.pretty:
    for test, duration, result in zip(tests, durations, results):
        print('%30s %10s %10s' % (test, duration, result))
else:
    maxnamelen = max([len(test) for test in tests])
    for test, duration in zip(tests, durations):
        comment = ''
        if duration > 1.0:
            if np.isfinite(duration):
                comment = '# ~%ds' % int(duration)
            else:
                comment = '# duration unknown'
        print(''.join(['    ',
                       ("'%s'," % test).ljust(maxnamelen + 5),
                       comment]).rstrip())



from __future__ import print_function
from gpaw import GPAW, restart, FermiDirac
from ase import Atoms
from gpaw.test import equal
import numpy as np

a = 5.404
bulk = Atoms(symbols='Si8',
             scaled_positions=[(0, 0, 0),
                               (0, 0.5, 0.5),
                               (0.5, 0, 0.5),
                               (0.5, 0.5, 0),
                               (0.25, 0.25, 0.25),
                               (0.25, 0.75, 0.75),
                               (0.75, 0.25, 0.75),
                               (0.75, 0.75, 0.25)],
             pbc=True, cell=(a, a, a))
n = 20
calc = GPAW(gpts=(n, n, n),
            nbands=8 * 3,
            occupations=FermiDirac(width=0.01),
            verbose=1,
            kpts=(1, 1, 1))
bulk.set_calculator(calc)
e1 = bulk.get_potential_energy()
niter1 = calc.get_number_of_iterations()
eigs = calc.get_eigenvalues(kpt=0)
calc.write('temp.gpw')

bulk, calc = restart('temp.gpw', fixdensity=True)
e2 = bulk.get_potential_energy()
eigs2 = calc.get_eigenvalues(kpt=0)
print('Orginal', eigs)
print('Fixdensity', eigs2)
print('Difference', eigs2 - eigs)

assert np.fabs(eigs2 - eigs)[:-1].max() < 3e-5
equal(e1, -36.767, 0.003)
equal(e1, e2, 1e-4)



from ase.build import bulk
from ase.dft.kpoints import monkhorst_pack
from gpaw import GPAW, FermiDirac
from gpaw.test import equal
from gpaw.xc.fxc import FXCCorrelation
from gpaw.mpi import world, serial_comm

if world.rank == 0:
    a0  = 5.43
    Ni = bulk('Ni', 'fcc')
    Ni.set_initial_magnetic_moments([0.7])

    kpts = monkhorst_pack((3,3,3))

    calc = GPAW(mode='pw',
                kpts=kpts,
                occupations=FermiDirac(0.001),
                setups={'Ni': '10'},
                communicator=serial_comm)
    
    Ni.set_calculator(calc)
    Ni.get_potential_energy()
    calc.diagonalize_full_hamiltonian()
    calc.write('Ni.gpw', mode='all')

world.barrier()

rpa = FXCCorrelation('Ni.gpw', xc='RPA',
                     nfrequencies=8, skip_gamma=True)
E_rpa = rpa.calculate(ecut=[50])

ralda = FXCCorrelation('Ni.gpw', xc='rALDA', unit_cells=[2,1,1],
                       nfrequencies=8, skip_gamma=True)
E_ralda = ralda.calculate(ecut=[50])

rapbe = FXCCorrelation('Ni.gpw', xc='rAPBE', unit_cells=[2,1,1],
                       nfrequencies=8, skip_gamma=True)
E_rapbe = rapbe.calculate(ecut=[50])

equal(E_rpa, -7.827, 0.01)
equal(E_ralda, -7.501, 0.01)
equal(E_rapbe, -7.444, 0.01)
